from __future__ import annotations

from django.contrib.auth.models import User
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.timezone import localtime


class Viewed(models.Model):
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, editable=False)
    object_id = models.PositiveIntegerField(editable=False)
    content_object = GenericForeignKey('content_type', 'object_id')

    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        editable=False,
        related_name='views',
        related_query_name='view'
    )

    created_datetime = models.DateTimeField(default=localtime, editable=False)

    def __str__(self):
        return f'{self.user} viewed {self.content_object} at {self.created_datetime}'

    class Meta:
        db_table = 'django_spire_history_viewed'
        verbose_name = 'Viewed'
        verbose_name_plural = 'Views'



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Page


class BasePage:
    def __init__(self, page: Page, base_url: str) -> None:
        self.base_url = base_url
        self.page = page

    def goto(self, url: str) -> None:
        self.page.goto(url)

    def has_text(self, text: str) -> bool:
        return text in self.page.locator('body').inner_text()

    def wait_for_load_state(self, state: str = 'networkidle') -> None:
        self.page.wait_for_load_state(state)

    def wait_for_timeout(self, timeout: int) -> None:
        self.page.wait_for_timeout(timeout)



from __future__ import annotations

from dandy import Prompt


def negation_prompt(
    negation_rules: list[str]
) -> Prompt:
    return (
        Prompt()
        .text('This step is mission critical. Do NOT deviate from this requirement.')
        .text('The following are strictly reserved and must never be used under any circumstances:')
        .list(negation_rules)
    )



from __future__ import annotations

from django.contrib import admin

from django_spire.help_desk.models import HelpDeskTicket


@admin.register(HelpDeskTicket)
class HelpDeskTicketAdmin(admin.ModelAdmin):
    list_display = ('pk', 'purpose', 'priority', 'status', 'created_by', 'created_datetime')
    list_filter = ('priority', 'purpose', 'status', 'is_active', 'is_deleted')
    ordering = ('-created_datetime',)
    raw_id_fields = ('created_by',)
    readonly_fields = ('created_by', 'created_datetime', 'is_active', 'is_deleted')
    search_fields = ('description', 'created_by__username', 'created_by__email')



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Locator, Page


class Toast:
    """Playwright component for django_spire/messages/messages.html"""

    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def container(self) -> Locator:
        return self.page.locator('.fixed-top[\\@notify\\.window]')

    @property
    def toasts(self) -> Locator:
        return self.container.locator('[x-data*="show:"]')

    def get_toast(self, index: int = 0) -> Locator:
        return self.toasts.nth(index)

    def get_toast_count(self) -> int:
        return self.toasts.count()

    def get_toast_icon(self, index: int = 0) -> Locator:
        return self.get_toast(index).locator('i.bi').first

    def get_toast_message(self, index: int = 0) -> str:
        return self.get_toast(index).locator('[x-text="notification.message"]').inner_text()

    def get_toast_type(self, index: int = 0) -> str:
        toast = self.get_toast(index)
        classes = toast.get_attribute('class') or ''

        if 'border-app-success' in classes:
            return 'success'

        if 'border-app-warning' in classes:
            return 'warning'

        if 'border-app-danger' in classes:
            return 'error'

        if 'border-app-primary' in classes:
            return 'info'

        return 'unknown'

    def close_toast(self, index: int = 0) -> None:
        self.get_toast(index).locator('.bi-x-lg').click()

    def has_success_toast(self) -> bool:
        return self.page.locator('.border-app-success').count() > 0

    def has_warning_toast(self) -> bool:
        return self.page.locator('.border-app-warning').count() > 0

    def has_error_toast(self) -> bool:
        return self.page.locator('.border-app-danger').count() > 0

    def has_info_toast(self) -> bool:
        return self.page.locator('.border-app-primary').count() > 0

    def wait_for_toast(self, timeout: int = 5000) -> None:
        self.toasts.first.wait_for(state='visible', timeout=timeout)

    def wait_for_toast_to_disappear(self, timeout: int = 10000) -> None:
        self.toasts.first.wait_for(state='hidden', timeout=timeout)



from __future__ import annotations

from django_spire.history.mixins import HistoryModelMixin


class PushNotification(HistoryModelMixin):
    class Meta:
        db_table = 'django_spire_notification_push'
        verbose_name = 'Push Notification'
        verbose_name_plural = 'Push Notifications'



from __future__ import annotations

import pytest

from django.test import TestCase
from django.utils import timezone

from django_spire.contrib.seeding.field.custom import CustomFieldSeeder


class TestCustomFieldSeeder(TestCase):
    def test_date_time_between_returns_aware_datetime(self) -> None:
        seeder = CustomFieldSeeder(fields={})

        result = seeder.date_time_between(start_date='-30d', end_date='now')

        assert timezone.is_aware(result)

    def test_in_order_cycles_through_values(self) -> None:
        seeder = CustomFieldSeeder(fields={})
        values = ['a', 'b', 'c']

        assert seeder.in_order(values, index=0) == 'a'
        assert seeder.in_order(values, index=1) == 'b'
        assert seeder.in_order(values, index=2) == 'c'
        assert seeder.in_order(values, index=3) == 'a'

    def test_in_order_raises_for_empty_list(self) -> None:
        seeder = CustomFieldSeeder(fields={})

        with pytest.raises(ValueError, match='Cannot select from empty values list'):
            seeder.in_order(values=[], index=0)

    def test_keyword_is_custom(self) -> None:
        assert CustomFieldSeeder.keyword == 'custom'



from __future__ import annotations

from django.urls import path

from django_spire.auth.group.views import json_views


app_name = 'group'

# Group
urlpatterns = [
    path('group/<int:pk>/permission/<str:app_name>/ajax/',
         json_views.permission_form_ajax,
         name='group_permission_ajax'),

    path('group/<int:pk>/app/<str:app_name>/special/role/ajax/',
         json_views.special_role_form_ajax,
         name='group_special_role_ajax'),
]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Locator, Page


class Modal:
    """Playwright component for django_spire/modal/modal.html"""

    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def close_button(self) -> Locator:
        return self.modal.locator('[\\@click*="close_modal"]').first

    @property
    def content(self) -> Locator:
        return self.modal.locator('.bg-app-layer-one.card')

    @property
    def modal(self) -> Locator:
        return self.page.locator('[x-show="show_modal"]').last

    @property
    def overlay(self) -> Locator:
        return self.page.locator('[\\@click="close_modal"]').first

    def close(self) -> None:
        self.close_button.click()

    def close_by_overlay(self) -> None:
        self.overlay.click()

    def is_open(self) -> bool:
        return self.modal.is_visible()

    def wait_for_close(self, timeout: int = 5000) -> None:
        self.modal.wait_for(state='hidden', timeout=timeout)

    def wait_for_open(self, timeout: int = 5000) -> None:
        self.modal.wait_for(state='visible', timeout=timeout)


class TitleModal(Modal):
    """Playwright component for django_spire/modal/title_modal.html"""

    @property
    def title(self) -> Locator:
        return self.modal.locator('.text-uppercase.text-app-secondary-dark.h6').first

    def get_title_text(self) -> str:
        return self.title.inner_text()


class FormModal(TitleModal):
    """Playwright component for modals containing forms"""

    @property
    def cancel_button(self) -> Locator:
        return self.modal.locator('button:has-text("Cancel"), [type="button"]:has-text("Cancel")').first

    @property
    def form(self) -> Locator:
        return self.modal.locator('form')

    @property
    def submit_button(self) -> Locator:
        return self.modal.locator('button[type="submit"], input[type="submit"]').first

    def cancel(self) -> None:
        self.cancel_button.click()

    def fill_field(self, name: str, value: str) -> None:
        self.form.locator(f'[name="{name}"]').fill(value)

    def get_field_value(self, name: str) -> str:
        return self.form.locator(f'[name="{name}"]').input_value()

    def submit(self) -> None:
        self.submit_button.click()


class DeleteModal(TitleModal):
    """Playwright component for delete confirmation modals"""

    @property
    def cancel_button(self) -> Locator:
        return self.modal.locator('button:has-text("Cancel")').first

    @property
    def confirm_button(self) -> Locator:
        return self.modal.locator('button:has-text("Delete"), button:has-text("Confirm")').first

    def cancel(self) -> None:
        self.cancel_button.click()

    def confirm(self) -> None:
        self.confirm_button.click()



from __future__ import annotations

from dandy import Bot

from django_spire.ai.prompt import prompts
from django_spire.ai.prompt import intel


class DandyPythonPromptBot(Bot):
    llm_role = prompts.dandy_prompt_python_file_instruction_bot_prompt()

    def process(self, prompt: str) -> intel.DandyPromptPythonFileIntel:
        return self.llm.prompt_to_intel(
            prompt=prompts.dandy_prompt_python_file_input_prompt(prompt),
            intel_class=intel.DandyPromptPythonFileIntel
        )


class TextToMarkdownPromptBot(Bot):
    llm_role = prompts.text_to_markdown_instruction_bot_prompt()

    def process(self, text: str) -> intel.TextToMarkdownIntel:
        return self.llm.prompt_to_intel(
            prompt=prompts.text_to_markdown_input_prompt(text),
            intel_class=intel.TextToMarkdownIntel
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib.contenttypes.fields import GenericRelation

from django_spire.comment.models import Comment
from django_spire.history.activity.mixins import ActivityMixin

if TYPE_CHECKING:
    from django.contrib.auth.models import User


class CommentModelMixin(ActivityMixin):
    comments = GenericRelation(
        Comment,
        related_query_name='comment',
        editable=False
    )

    def add_comment(
            self,
            user: User,
            information: str,
            parent: int | Comment | None = None
    ):
        if isinstance(parent, Comment):
            parent = parent.pk

        return self.comments.create(
            information=information,
            user=user,
            parent_id=parent
        )

    class Meta:
        abstract = True



from __future__ import annotations

from datetime import datetime
from unittest.mock import MagicMock

from django.contrib.sessions.middleware import SessionMiddleware
from django.contrib.sessions.models import Session
from django.test import RequestFactory, TestCase

from django_spire.contrib.session.controller import SessionController
from django_spire.contrib.session.templatetags.session_tags import session_controller_to_json


class TestSessionController(TestCase):
    def setUp(self) -> None:
        self.factory = RequestFactory()
        self.request = self.factory.get('/')

        middleware = SessionMiddleware(lambda req: None)
        middleware.process_request(self.request)

        self.shopping_cart_session = SessionController(
            request=self.request,
            session_key='shopping_cart'
        )

    def test_add_data(self) -> None:
        key = 'currency'
        value = 'CAD'

        self.shopping_cart_session.add_data(key, value)

        assert key in self.shopping_cart_session.data
        assert self.shopping_cart_session.data[key] == value

    def test_data_property_returns_dict(self) -> None:
        assert isinstance(self.shopping_cart_session.data, dict)

    def test_getitem(self) -> None:
        self.shopping_cart_session.add_data('currency', 'CAD')

        assert self.shopping_cart_session['currency'] == 'CAD'

    def test_has_data_false_when_empty(self) -> None:
        assert self.shopping_cart_session.has_data is False

    def test_has_data_true_when_data_exists(self) -> None:
        self.shopping_cart_session.add_data('currency', 'CAD')

        assert self.shopping_cart_session.has_data is True

    def test_has_data_false_after_remove(self) -> None:
        self.shopping_cart_session.add_data('currency', 'CAD')
        self.shopping_cart_session.remove_data('currency')

        assert self.shopping_cart_session.has_data is False

    def test_is_expired_false_when_not_expired(self) -> None:
        self.shopping_cart_session.add_data('currency', 'CAD')

        assert self.shopping_cart_session.is_expired is False

    def test_is_expired_true_when_expired(self) -> None:
        self.shopping_cart_session.add_data('currency', 'CAD')
        self.shopping_cart_session.seconds_till_expiry = -5
        self.shopping_cart_session._set_timeout_datestamp()

        assert self.shopping_cart_session.is_expired is True

    def test_purge_removes_session_data(self) -> None:
        self.shopping_cart_session.add_data('currency', 'CAD')
        self.shopping_cart_session.purge()

        assert 'shopping_cart' not in self.request.session

    def test_remove_data(self) -> None:
        self.shopping_cart_session.add_data('currency', 'CAD')
        self.shopping_cart_session.add_data('discount', None)
        self.shopping_cart_session.remove_data('currency')

        assert 'currency' not in self.shopping_cart_session.data
        assert 'discount' in self.shopping_cart_session.data

    def test_session_data_added_to_database(self) -> None:
        self.shopping_cart_session.add_data('favorites', ['cool_shirt', 'fun_hat'])
        self.request.session.save()

        session_key = self.request.session.session_key
        session_in_db = Session.objects.get(session_key=session_key)
        session_data = session_in_db.get_decoded()

        assert self.shopping_cart_session.session_key in session_data
        assert 'favorites' in session_data[self.shopping_cart_session.session_key]
        assert session_data[self.shopping_cart_session.session_key]['favorites'] == ['cool_shirt', 'fun_hat']

    def test_set_modified(self) -> None:
        self.shopping_cart_session.add_data('currency', 'CAD')

        assert self.request.session.modified is True

    def test_set_timeout_datestamp(self) -> None:
        self.shopping_cart_session._set_timeout_datestamp()

        assert '_timeout_datestamp' in self.shopping_cart_session.data
        assert datetime.now().timestamp() < self.shopping_cart_session['_timeout_datestamp']

    def test_setitem(self) -> None:
        key = 'currency'
        value = 'CAD'

        self.shopping_cart_session[key] = value

        assert key in self.shopping_cart_session.data
        assert self.shopping_cart_session.data[key] == value

    def test_timeout_datestamp_returns_float(self) -> None:
        self.shopping_cart_session._set_timeout_datestamp()

        assert isinstance(self.shopping_cart_session.timeout_datestamp, float)

    def test_timeout_datestamp_returns_zero_when_not_set(self) -> None:
        controller = SessionController(
            request=self.request,
            session_key='new_session'
        )

        assert controller.timeout_datestamp == 0

    def test_to_json_returns_string(self) -> None:
        self.shopping_cart_session.add_data('currency', 'CAD')
        self.shopping_cart_session.add_data('discount', None)

        assert isinstance(self.shopping_cart_session.to_json(), str)


class TestSessionControllerToJson(TestCase):
    def setUp(self) -> None:
        self.factory = RequestFactory()

    def _create_context_with_session(self) -> MagicMock:
        request = self.factory.get('/')
        middleware = SessionMiddleware(lambda req: None)
        middleware.process_request(request)

        context = MagicMock()
        context.get.return_value = request

        return context

    def test_returns_escaped_json(self) -> None:
        context = self._create_context_with_session()

        result = session_controller_to_json(context, 'test_key')

        assert isinstance(result, str)

    def test_calls_context_get_with_request(self) -> None:
        context = self._create_context_with_session()

        session_controller_to_json(context, 'test_key')

        context.get.assert_called_once_with('request')

    def test_returns_safe_string(self) -> None:
        context = self._create_context_with_session()

        result = session_controller_to_json(context, 'test_key')

        assert hasattr(result, '__html__')



from __future__ import annotations

import pytest

from unittest.mock import MagicMock, patch

from django.http import HttpRequest, JsonResponse
from django.test import RequestFactory, TestCase

from django_spire.core.decorators import close_db_connections, valid_ajax_request_required


class TestCloseDbConnections(TestCase):
    @patch('django_spire.core.decorators.connections')
    def test_closes_connections_after_exception(self, mock_connections: MagicMock) -> None:
        @close_db_connections
        def sample_func():
            raise ValueError

        with pytest.raises(ValueError):
            sample_func()

        mock_connections.close_all.assert_called_once()

    @patch('django_spire.core.decorators.connections')
    def test_closes_connections_after_success(self, mock_connections: MagicMock) -> None:
        @close_db_connections
        def sample_func():
            return 'result'

        result = sample_func()

        assert result == 'result'
        mock_connections.close_all.assert_called_once()

    @patch('django_spire.core.decorators.connections')
    def test_preserves_function_metadata(self, _mock_connections: MagicMock) -> None:
        @close_db_connections
        def sample_func():
            """Sample docstring"""

        assert sample_func.__name__ == 'sample_func'
        assert sample_func.__doc__ == 'Sample docstring'


class TestValidAjaxRequestRequired(TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.factory = RequestFactory()

    def test_get_request_returns_error(self) -> None:
        @valid_ajax_request_required
        def sample_view(_request: HttpRequest) -> dict:
            return {'success': True}

        request = self.factory.get('/test/')
        response = sample_view(request)

        assert response.status_code == 200
        assert b'error' in response.content

    def test_post_request_with_json_content_type_succeeds(self) -> None:
        @valid_ajax_request_required
        def sample_view(_request: HttpRequest) -> JsonResponse:
            return JsonResponse({'success': True})

        request = self.factory.post(
            '/test/',
            data='{}',
            content_type='application/json'
        )
        response = sample_view(request)

        assert response.status_code == 200
        assert b'success' in response.content

    def test_post_request_without_json_content_type_passes(self) -> None:
        @valid_ajax_request_required
        def sample_view(_request: HttpRequest) -> JsonResponse:
            return JsonResponse({'success': True})

        request = self.factory.post('/test/', data={})
        response = sample_view(request)

        assert response.status_code == 200
        assert b'success' in response.content

    def test_preserves_function_metadata(self) -> None:
        @valid_ajax_request_required
        def sample_view(_request: HttpRequest) -> None:
            """Sample docstring"""

        assert sample_view.__name__ == 'sample_view'
        assert sample_view.__doc__ == 'Sample docstring'



from django_spire.testing.playwright.components import (
    Accordion,
    AttributeElement,
    AttributeList,
    Breadcrumb,
    Card,
    DeleteModal,
    Dropdown,
    EllipsisDropdown,
    EllipsisModalDropdown,
    EllipsisTableDropdown,
    FilterForm,
    FormCard,
    FormModal,
    InfiniteScroll,
    InfiniteScrollCard,
    InfiniteScrollList,
    InfiniteScrollTable,
    LazyTab,
    Modal,
    NavAccordion,
    NotificationBell,
    SideNavigation,
    ThemeSelector,
    TitleCard,
    TitleModal,
    Toast,
    TopNavigation,
    UserMenu,
)
from django_spire.testing.playwright.pages import BasePage


__all__ = [
    'Accordion',
    'AttributeElement',
    'AttributeList',
    'BasePage',
    'Breadcrumb',
    'Card',
    'DeleteModal',
    'Dropdown',
    'EllipsisDropdown',
    'EllipsisModalDropdown',
    'EllipsisTableDropdown',
    'FilterForm',
    'FormCard',
    'FormModal',
    'InfiniteScroll',
    'InfiniteScrollCard',
    'InfiniteScrollList',
    'InfiniteScrollTable',
    'LazyTab',
    'Modal',
    'NavAccordion',
    'NotificationBell',
    'SideNavigation',
    'ThemeSelector',
    'TitleCard',
    'TitleModal',
    'Toast',
    'TopNavigation',
    'UserMenu',
]



from __future__ import annotations

from dandy import Prompt


def prompt_tuning_instruction_bot_prompt():
    return (
        Prompt()
        .heading('System Prompt Tuning Expert')
        .text(
            'You are an expert prompt engineer specializing in analyzing and tuning system prompts for Large Language Models (LLMs).')
        .line_break()
        .heading('Goal').text('Analyze the provided prompt and user feedback to create a tuned version of the system prompt that meets the users\' needs.')
        .line_break()
        .heading('Rules')
        .ordered_list([
            'Preserve Core Purpose: Maintain the fundamental goal and role of the original prompt while making improvements.',
            'Small Incremental Improvements: Focus on small changes as the user will keep providing more feedback until they are satisfied.',
            'Adaptability: You can add, create, or remove parts of the prompt to better align with the user\'s feedback but ensure to keep the overall structure. ',
        ])
        .line_break()
        .heading('Input Format')
        .text('You will receive:')
        .ordered_list([
            'System Prompt: The original system prompt that needs tuning',
            'Feedback: The user\'s feedback about issues with the current prompt and what they want to improve',
        ])
    )


def formatting_bot_instruction_prompt():
    return (
        Prompt()
        .heading('Prompt Formatting Expert')
        .text(
            'You are a specialized prompt engineer focused on improving the formatting and structure of system prompts for Large Language Models (LLMs).')
        .line_break()
        .heading('Goal')
        .text(
            'Analyze the provided system prompt and improve its formatting and structure while preserving its content and intent.')
        .line_break()
        .heading('Rules')
        .ordered_list([
            'Preserve Content: Maintain all the information and instructions from the original prompt.',
            'Improve Structure: Organize content into clear sections with appropriate headings.',
            'Standardize Formatting: Apply consistent formatting for similar elements (lists, examples, etc.).',
            'Enhance Readability: Use proper spacing, indentation, and line breaks to improve readability.',
            'Maintain Hierarchy: Ensure logical hierarchy of information with primary and secondary points clearly distinguished.',
        ])
        .line_break()
        .heading('Input Format')
        .text('You will receive:')
        .ordered_list([
            'System Prompt: The original system prompt that needs formatting improvements',
        ])
        .line_break()
        .heading('Output Format')
        .text('Return a reformatted system prompt that maintains the original content but with improved structure and formatting.')
    )


def instruction_clarity_bot_instruction_prompt():
    return (
        Prompt()
        .heading('Instruction Clarity Expert')
        .text(
            'You are a specialized prompt engineer focused on improving the clarity of instructions in system prompts for Large Language Models (LLMs).')
        .line_break()
        .heading('Goal')
        .text(
            'Analyze the provided system prompt and improve the clarity of its instructions while maintaining its intent.')
        .line_break()
        .heading('Rules')
        .ordered_list([
            'Identify Ambiguities: Find and resolve any ambiguous or unclear instructions.',
            'Enhance Precision: Make instructions more precise and specific where needed.',
            'Improve Logical Flow: Ensure instructions follow a logical sequence.',
            'Add Context: Provide necessary context for complex instructions.',
            'Simplify Complex Instructions: Break down complex instructions into simpler steps when appropriate.',
        ])
        .line_break()
        .heading('Input Format')
        .text('You will receive:')
        .ordered_list([
            'System Prompt: The original system prompt that needs instruction clarity improvements',
        ])
        .line_break()
        .heading('Output Format')
        .text('Return an improved system prompt with clearer, more precise instructions.')
    )


def persona_bot_instruction_prompt():
    return (
        Prompt()
        .heading('Persona Consistency Expert')
        .text(
            'You are a specialized prompt engineer focused on maintaining consistent tone and persona in system prompts for Large Language Models (LLMs).')
        .line_break()
        .heading('Goal')
        .text(
            'Analyze the provided system prompt and ensure it maintains a consistent tone, voice, and persona throughout.')
        .line_break()
        .heading('Rules')
        .ordered_list([
            'Identify Tone: Determine the intended tone of the prompt (formal, conversational, authoritative, etc.).',
            'Ensure Consistency: Make sure the tone and voice remain consistent throughout the prompt.',
            'Align with Purpose: Ensure the persona aligns with the prompt\'s purpose and intended audience.',
            'Standardize Language: Use consistent terminology and phrasing throughout.',
            'Preserve Character: If the prompt defines a specific character or role, ensure all language aligns with that character.',
        ])
        .line_break()
        .heading('Input Format')
        .text('You will receive:')
        .ordered_list([
            'System Prompt: The original system prompt that needs persona consistency improvements',
        ])
        .line_break()
        .heading('Output Format')
        .text('Return an improved system prompt with consistent tone, voice, and persona throughout.')
    )


def duplication_removal_bot_instruction_prompt():
    return (
        Prompt()
        .heading('Duplication Removal Expert')
        .text(
            'You are a specialized prompt engineer focused on identifying and removing redundancies in system prompts for Large Language Models (LLMs).')
        .line_break()
        .heading('Goal')
        .text(
            'Analyze the provided system prompt and remove any duplicated or redundant content while preserving all unique information.')
        .line_break()
        .heading('Rules')
        .ordered_list([
            'Identify Redundancies: Find repeated information, instructions, or concepts.',
            'Consolidate Information: Combine similar or related points into concise statements.',
            'Preserve Unique Content: Ensure no unique information is lost when removing duplications.',
            'Maintain Context: Keep necessary context when consolidating information.',
            'Improve Flow: Ensure the prompt flows naturally after removing redundancies.',
        ])
        .line_break()
        .heading('Input Format')
        .text('You will receive:')
        .ordered_list([
            'System Prompt: The original system prompt that may contain duplicated content',
        ])
        .line_break()
        .heading('Output Format')
        .text('Return an improved system prompt with duplications removed and information consolidated.')
    )


def example_optimization_bot_instruction_prompt():
    return (
        Prompt()
        .heading('Example Optimization Expert')
        .text(
            'You are a specialized prompt engineer focused on refining examples within system prompts for Large Language Models (LLMs).')
        .line_break()
        .heading('Goal')
        .text(
            'Analyze the provided system prompt and optimize the examples to better illustrate the desired behavior and outputs.')
        .line_break()
        .heading('Rules')
        .ordered_list([
            'Assess Current Examples: Evaluate the effectiveness of existing examples.',
            'Improve Clarity: Make examples clearer and more illustrative of the desired behavior.',
            'Ensure Diversity: Ensure examples cover a range of scenarios and edge cases.',
            'Balance Complexity: Include both simple and more complex examples when appropriate.',
            'Align with Instructions: Ensure examples clearly demonstrate the instructions they support.',
        ])
        .line_break()
        .heading('Input Format')
        .text('You will receive:')
        .ordered_list([
            'System Prompt: The original system prompt containing examples that need optimization',
        ])
        .line_break()
        .heading('Output Format')
        .text('Return an improved system prompt with optimized examples that better illustrate the desired behavior.')
    )


def prompt_tuning_input_prompt(system_prompt: str, feedback: str):
    return (
        Prompt()
        .heading('User Supplies System Prompt')
        .line_break()
        .text(system_prompt)
        .line_break()
        .heading('Feedback')
        .text(feedback)
    )


def specialized_bot_input_prompt(system_prompt: str):
    return (
        Prompt()
        .heading('System Prompt for Analysis')
        .line_break()
        .text(system_prompt)
    )



from __future__ import annotations

from django.urls import path

from django_spire.knowledge.collection.views import form_views


app_name = 'form'

urlpatterns = [
    path('create/', form_views.form_view, name='create'),
    path('<int:parent_pk>/create_with_parent/', form_views.form_view, name='create_with_parent'),
    path('<int:pk>/update/', form_views.form_view, name='update'),
]



from django_spire.metric.report.enums import ColumnType


def get_text_alignment_css_class(column_type: ColumnType) -> str:
    if column_type in (
            ColumnType.DOLLAR,
            ColumnType.DOLLAR_1,
            ColumnType.DOLLAR_2,
            ColumnType.DOLLAR_3,
            ColumnType.PERCENT,
            ColumnType.PERCENT_1,
            ColumnType.PERCENT_2,
            ColumnType.PERCENT_3,
            ColumnType.NUMBER,
            ColumnType.NUMBER_1,
            ColumnType.NUMBER_2,
            ColumnType.NUMBER_3
    ):
        return 'text-end'
    if column_type == ColumnType.CHOICE:
        return 'text-center'

    return 'text-start'



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.contrib.seeding.field.cleaners import normalize_seeder_fields
from django_spire.contrib.seeding.model.enums import ModelSeederDefaultsEnum

if TYPE_CHECKING:
    from typing import Any, Self


class FieldsConfig:
    def __init__(
        self,
        raw_fields: dict[str, Any],
        field_names: list[str],
        default_to: str,
        model_class: Any
    ) -> None:
        self.raw_fields = raw_fields or {}
        self.default_to = default_to
        self.field_names = set(field_names)
        self.model_class = model_class

        self._excluded = {
            k for k, v in self.raw_fields.items()
            if v in ("exclude", ("exclude",))
        }

        self.fields = normalize_seeder_fields({
            k: v for k, v in self.raw_fields.items() if k not in self._excluded
        })

        self._validate()
        self._assign_defaults()

        # Fields need to be in order for caching hash
        self._order_fields()

    def _validate(self) -> None:
        unknown = set(self.fields.keys()) - self.field_names

        if unknown:
            message = f"Invalid field name(s): {', '.join(unknown)}"
            raise ValueError(message)

    def _assign_defaults(self) -> None:
        if self.default_to == ModelSeederDefaultsEnum.INCLUDED:
            return

        method = self.default_to.lower()

        for name in self.field_names:
            if name not in self.fields and name not in self._excluded:
                self.fields[name] = (method,)

    def override(self, overrides: dict) -> Self:
        merged = {**self.fields, **normalize_seeder_fields(overrides)}
        new_raw = {
            **{k: v for k, v in self.raw_fields.items() if k in self._excluded},
            **merged
        }
        return self.__class__(
            raw_fields=new_raw,
            field_names=list(self.field_names),
            default_to=self.default_to,
            model_class=self.model_class
        )

    def _order_fields(self) -> None:
        self.fields = dict(
            sorted(
                self.fields.items(),
                key=lambda x: x[0]
            )
        )

    @property
    def excluded(self) -> set:
        return self._excluded



from __future__ import annotations

from typing import TYPE_CHECKING

from django.http import HttpResponse

from django_spire.ai.chat.models import Chat
from django_spire.ai.chat.responses import MessageResponseGroup

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


def load_messages_render_view(request: WSGIRequest, chat_id: int) -> HttpResponse:
    chat = (
        Chat.objects
        .by_user(request.user)
        .get(id=chat_id)
    )

    message_group = MessageResponseGroup()

    for chat_message in chat.messages.newest_by_count_reversed(20):
        message_group.add_message_response(
            chat_message.to_message_response()
        )

    return HttpResponse(
        message_group.render_to_html_string(
            {
                'chat_id': chat.id,
                'is_loading': True,
            }
        )
    )



from __future__ import annotations

import pytest

from abc import ABC

from django.contrib.auth.models import User
from django.test import TestCase

from django_spire.contrib.constructor import (
    BaseConstructor,
    BaseDjangoModelConstructor,
    ConstructorError,
)


class TestBaseConstructor(TestCase):
    def test_abc_subclass_does_not_require_obj_annotation(self) -> None:
        class AbstractConstructor(BaseConstructor, ABC):
            pass

        # Should not raise
        assert AbstractConstructor is not None

    def test_constructor_with_none_obj(self) -> None:
        class StringConstructor(BaseConstructor[str]):
            obj: str

        constructor = StringConstructor(None)

        assert not hasattr(constructor, 'obj')

    def test_constructor_with_valid_obj(self) -> None:
        class StringConstructor(BaseConstructor[str]):
            obj: str

        constructor = StringConstructor('test')

        assert constructor.obj == 'test'

    def test_constructor_with_wrong_type_raises_error(self) -> None:
        class StringConstructor(BaseConstructor[str]):
            obj: str

        with pytest.raises(ConstructorError):
            StringConstructor(123)

    def test_obj_class_property(self) -> None:
        class StringConstructor(BaseConstructor[str]):
            obj: str

        constructor = StringConstructor('test')

        assert constructor.obj_class is str

    def test_obj_is_valid_property(self) -> None:
        class StringConstructor(BaseConstructor[str]):
            obj: str

        constructor = StringConstructor('test')

        assert constructor._obj_is_valid is True

    def test_post_init_called(self) -> None:
        class StringConstructor(BaseConstructor[str]):
            obj: str
            post_init_called: bool = False

            def __post_init__(self):
                self.post_init_called = True

        constructor = StringConstructor('test')

        assert constructor.post_init_called is True

    def test_subclass_without_obj_annotation_raises_error(self) -> None:
        with pytest.raises(ConstructorError, match='must have an "obj" attribute'):
            class InvalidConstructor(BaseConstructor[str]):
                pass


class TestBaseDjangoModelConstructor(TestCase):
    def setUp(self) -> None:
        self.user = User.objects.create_user(
            username='testuser',
            password='testpass'  # noqa: S106
        )

    def test_constructor_with_django_model(self) -> None:
        class UserConstructor(BaseDjangoModelConstructor[User]):
            obj: User

        constructor = UserConstructor(self.user)

        assert constructor.obj == self.user

    def test_model_obj_is_created_true(self) -> None:
        class UserConstructor(BaseDjangoModelConstructor[User]):
            obj: User

        constructor = UserConstructor(self.user)

        assert constructor.model_obj_is_created is True

    def test_model_obj_is_created_false_for_new(self) -> None:
        class UserConstructor(BaseDjangoModelConstructor[User]):
            obj: User

        new_user = User(username='newuser')
        constructor = UserConstructor(new_user)

        assert constructor.model_obj_is_created is False

    def test_model_obj_is_new_false(self) -> None:
        class UserConstructor(BaseDjangoModelConstructor[User]):
            obj: User

        constructor = UserConstructor(self.user)

        assert constructor.model_obj_is_new is False

    def test_model_obj_is_new_true(self) -> None:
        class UserConstructor(BaseDjangoModelConstructor[User]):
            obj: User

        new_user = User(username='newuser')
        constructor = UserConstructor(new_user)

        assert constructor.model_obj_is_new is True

    def test_model_obj_pk_is_empty_true(self) -> None:
        class UserConstructor(BaseDjangoModelConstructor[User]):
            obj: User

        new_user = User(username='newuser')
        constructor = UserConstructor(new_user)

        assert constructor._model_obj_pk_is_empty is True

    def test_model_obj_pk_is_empty_false(self) -> None:
        class UserConstructor(BaseDjangoModelConstructor[User]):
            obj: User

        constructor = UserConstructor(self.user)

        assert constructor._model_obj_pk_is_empty is False

    def test_obj_class_returns_model_class(self) -> None:
        class UserConstructor(BaseDjangoModelConstructor[User]):
            obj: User

        constructor = UserConstructor(self.user)

        assert constructor.obj_class is User

    def test_obj_is_valid_true(self) -> None:
        class UserConstructor(BaseDjangoModelConstructor[User]):
            obj: User

        constructor = UserConstructor(self.user)

        assert constructor._obj_is_valid is True


class TestConstructorError(TestCase):
    def test_constructor_error_is_exception(self) -> None:
        assert issubclass(ConstructorError, Exception)

    def test_constructor_error_message(self) -> None:
        error = ConstructorError('Test error message')

        assert str(error) == 'Test error message'

    def test_constructor_error_can_be_raised(self) -> None:
        with pytest.raises(ConstructorError):
            message = 'Test'
            raise ConstructorError(message)


class TestConstructorDescriptor(TestCase):
    def setUp(self) -> None:
        self.user = User.objects.create_user(
            username='testuser',
            password='testpass'  # noqa: S106
        )

    def test_descriptor_get_from_model_instance(self) -> None:
        class UserConstructor(BaseDjangoModelConstructor[User]):
            obj: User

        constructor = UserConstructor(self.user)

        assert constructor.obj == self.user



from __future__ import annotations

from django.conf import settings as django_settings

from django_spire import settings as django_spire_default_settings


class Settings:
    def __getattr__(self, name: str):
        django_value = None
        django_spire_value = None

        if hasattr(django_settings, name):
            django_value = getattr(django_settings, name)

        if hasattr(django_spire_default_settings, name):
            django_spire_value = getattr(django_spire_default_settings, name)

        if name == 'DJANGO_SPIRE_AUTH_CONTROLLERS' and isinstance(django_value, dict) and isinstance(django_spire_value, dict):
            return {
                **django_spire_value,
                **django_value
            }

        if django_value is not None:
            return django_value

        if django_spire_value is not None:
            return django_spire_value

        return None


settings = Settings()



from __future__ import annotations

from unittest.mock import patch

from django.test import TestCase, override_settings

from django_spire.contrib.performance.decorators import performance_timer


class TestPerformanceTimer(TestCase):
    @override_settings(DEBUG=True)
    def test_calls_function_when_debug_true(self) -> None:
        @performance_timer
        def sample_func() -> str:
            return 'result'

        result = sample_func()

        assert result == 'result'

    @override_settings(DEBUG=False)
    def test_calls_function_when_debug_false(self) -> None:
        @performance_timer
        def sample_func() -> str:
            return 'result'

        result = sample_func()

        assert result == 'result'

    @override_settings(DEBUG=True)
    def test_logs_warning_when_debug_true(self) -> None:
        @performance_timer
        def sample_func() -> str:
            return 'result'

        with patch('django_spire.contrib.performance.decorators.log.warning') as mock_warning:
            sample_func()

            mock_warning.assert_called_once()

    @override_settings(DEBUG=False)
    def test_no_logging_when_debug_false(self) -> None:
        @performance_timer
        def sample_func() -> str:
            return 'result'

        with patch('django_spire.contrib.performance.decorators.log.warning') as mock_warning:
            sample_func()

            mock_warning.assert_not_called()

    @override_settings(DEBUG=True)
    def test_log_message_contains_runtime(self) -> None:
        @performance_timer
        def sample_func() -> str:
            return 'result'

        with patch('django_spire.contrib.performance.decorators.log.warning') as mock_warning:
            sample_func()

            log_message = mock_warning.call_args[0][0]

            assert 'seconds runtime' in log_message

    @override_settings(DEBUG=True)
    def test_log_message_contains_function_name(self) -> None:
        @performance_timer
        def sample_func() -> str:
            return 'result'

        with patch('django_spire.contrib.performance.decorators.log.warning') as mock_warning:
            sample_func()

            log_message = mock_warning.call_args[0][0]

            assert 'sample_func' in log_message

    @override_settings(DEBUG=True)
    def test_passes_args_to_function(self) -> None:
        @performance_timer
        def sample_func(a: int, b: int) -> int:
            return a + b

        result = sample_func(2, 3)

        assert result == 5

    @override_settings(DEBUG=True)
    def test_passes_kwargs_to_function(self) -> None:
        @performance_timer
        def sample_func(a: int, b: int = 10) -> int:
            return a + b

        result = sample_func(5, b=20)

        assert result == 25

    @override_settings(DEBUG=True)
    def test_returns_none_when_function_returns_none(self) -> None:
        @performance_timer
        def sample_func() -> None:
            pass

        result = sample_func()

        assert result is None



from __future__ import annotations

from django.test import override_settings

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.file.models import File
from django_spire.file.tests.factories import create_test_file


STORAGES_OVERRIDE = {
    'default': {
        'BACKEND': 'django.core.files.storage.FileSystemStorage',
    },
    'staticfiles': {
        'BACKEND': 'django.contrib.staticfiles.storage.StaticFilesStorage',
    },
}


@override_settings(STORAGES=STORAGES_OVERRIDE)
class FileQuerySetTests(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.file1 = create_test_file(name='File 1')
        self.file2 = create_test_file(name='File 2')

    def test_active_returns_non_deleted(self):
        result = File.objects.active()

        assert self.file1 in result
        assert self.file2 in result

    def test_active_excludes_deleted(self):
        self.file1.is_deleted = True
        self.file1.save()

        result = File.objects.active()

        assert self.file1 not in result
        assert self.file2 in result

    def test_active_excludes_inactive(self):
        self.file1.is_active = False
        self.file1.save()

        result = File.objects.active()

        assert self.file1 not in result
        assert self.file2 in result

    def test_active_excludes_deleted_and_inactive(self):
        self.file1.is_deleted = True
        self.file1.is_active = False
        self.file1.save()

        result = File.objects.active()

        assert self.file1 not in result
        assert self.file2 in result

    def test_related_field_filters_correctly(self):
        self.file1.related_field = 'abc'
        self.file1.save()
        self.file2.related_field = 'xyz'
        self.file2.save()

        result = File.objects.related_field('abc')

        assert self.file1 in result
        assert self.file2 not in result

    def test_related_field_returns_empty_when_no_match(self):
        self.file1.related_field = 'abc'
        self.file1.save()

        result = File.objects.related_field('xyz')

        assert result.count() == 0

    def test_related_field_filters_null_values(self):
        self.file1.related_field = None
        self.file1.save()
        self.file2.related_field = 'abc'
        self.file2.save()

        result = File.objects.related_field(None)

        assert self.file1 in result
        assert self.file2 not in result

    def test_chaining_active_and_related_field(self):
        self.file1.related_field = 'abc'
        self.file1.save()
        self.file2.related_field = 'abc'
        self.file2.is_deleted = True
        self.file2.save()

        result = File.objects.active().related_field('abc')

        assert self.file1 in result
        assert self.file2 not in result



from __future__ import annotations

from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from django_spire.history.viewed.models import Viewed


class TestViewed(TestCase):
    def setUp(self) -> None:
        self.user = User.objects.create_user(
            username='testuser',
            password='testpass'  # noqa: S106
        )
        self.content_type = ContentType.objects.get_for_model(User)

    def test_create_viewed(self) -> None:
        viewed = Viewed.objects.create(
            content_type=self.content_type,
            object_id=self.user.pk,
            user=self.user
        )

        assert viewed.pk is not None
        assert viewed.user == self.user
        assert viewed.content_object == self.user

    def test_created_datetime_auto_set(self) -> None:
        viewed = Viewed.objects.create(
            content_type=self.content_type,
            object_id=self.user.pk,
            user=self.user
        )

        assert viewed.created_datetime is not None

    def test_str_representation(self) -> None:
        viewed = Viewed.objects.create(
            content_type=self.content_type,
            object_id=self.user.pk,
            user=self.user
        )

        assert 'testuser' in str(viewed)
        assert 'viewed' in str(viewed)



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.help_desk.enums import TicketEventType


class TicketEventTypeTests(BaseTestCase):
    def test_new_value(self):
        assert TicketEventType.NEW.value == 'new'

    def test_update_value(self):
        assert TicketEventType.UPDATE.value == 'update'

    def test_comment_value(self):
        assert TicketEventType.COMMENT.value == 'comment'

    def test_enum_members_count(self):
        assert len(TicketEventType) == 3



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any

    from django_spire.core.management.commands.spire_startapp_pkg.reporter import Reporter


class PermissionInheritanceHandler:
    """
    Handles permission inheritance configuration for nested Django apps.

    This class manages the interactive collection of permission inheritance
    settings when creating child apps that may inherit permissions from parent apps.
    """

    def __init__(self, reporter: Reporter):
        """
        Initializes the handler with a reporter for user interaction.

        :param reporter: Reporter instance for displaying prompts and messages.
        """

        self.reporter = reporter

    def collect_inheritance_data(self, components: list[str]) -> dict[str, Any]:
        """
        Collects permission inheritance configuration from the user.

        Prompts the user to determine if the new app should inherit permissions
        from its parent app, and if so, collects the necessary parent model information.

        :param components: List of app path components.
        :return: Dictionary containing inheritance configuration data.
        """

        if len(components) <= 2:
            return {
                'inherit_permissions': False,
                'parent_model_instance_name': '',
                'parent_permission_prefix': '',
            }

        if not self._should_inherit_permissions():
            return {
                'inherit_permissions': False,
                'parent_model_instance_name': '',
                'parent_permission_prefix': '',
            }

        return {
            'inherit_permissions': True,
            'parent_model_instance_name': self._collect_parent_model_instance_name(components),
            'parent_model_path': self._collect_parent_model_path(components),
            'parent_permission_prefix': self._collect_parent_permission_prefix(components),
        }

    def _build_default_parent_model_path(self, components: list[str]) -> str:
        """
        Builds a default parent model path based on app components.

        :param components: List of app path components.
        :return: Default parent model path string.
        """

        parent_name = components[-2]
        parent_model_class = ''.join(word.title() for word in parent_name.split('_'))
        return '.'.join(components[:-1]) + f'.models.{parent_model_class}'

    def _build_default_parent_permission_prefix(self, components: list[str]) -> str:
        """
        Builds a default parent permission prefix based on app components.

        :param components: List of app path components.
        :return: Default permission prefix string.
        """

        parent_parts = components[1:-1]
        return '_'.join(parent_parts).lower()

    def _collect_parent_model_instance_name(self, components: list[str]) -> str:
        """
        Prompts the user for the parent model instance name.

        :param components: List of app path components.
        :return: User-provided or default parent model instance name.
        """

        parent_name = components[-2]
        default = parent_name.lower()

        self.reporter.write(
            f'\nEnter the parent model instance name (default: "{default}")',
            self.reporter.style_notice
        )

        user_input = input('Press Enter to use default or type a custom name: ').strip()
        return user_input if user_input else default

    def _collect_parent_model_path(self, components: list[str]) -> str:
        """
        Prompts the user for the parent model path.

        :param components: List of app path components.
        :return: User-provided or default parent model path.
        """

        default = self._build_default_parent_model_path(components)

        self.reporter.write(
            f'\nEnter the parent model path (default: "{default}")',
            self.reporter.style_notice
        )

        user_input = input('Press Enter to use default or type a custom path: ').strip()
        return user_input if user_input else default

    def _collect_parent_permission_prefix(self, components: list[str]) -> str:
        """
        Prompts the user for the parent permission prefix.

        :param components: List of app path components.
        :return: User-provided or default parent permission prefix.
        """

        default = self._build_default_parent_permission_prefix(components)

        self.reporter.write(
            f'\nEnter the parent permission prefix (default: "{default}")',
            self.reporter.style_notice
        )

        user_input = input('Press Enter to use default or type a custom prefix: ').strip()
        return user_input if user_input else default

    def _should_inherit_permissions(self) -> bool:
        """
        Prompts the user to confirm permission inheritance.

        :return: True if user wants to inherit permissions, False otherwise.
        """

        self.reporter.write('\nDo you want this app to inherit permissions from its parent? (y/n)', self.reporter.style_notice)
        user_input = input('Default is "n": ').strip().lower()
        return user_input == 'y'



from __future__ import annotations

import traceback

from typing import TYPE_CHECKING

from django_spire.contrib.service import BaseDjangoModelService
from django_spire.core.decorators import close_db_connections
from django_spire.knowledge.entry.version.block.models import EntryVersionBlock
from django_spire.knowledge.exceptions import KnowledgeBaseConversionError

if TYPE_CHECKING:
    from django_spire.knowledge.entry.models import Entry


class EntryAutomationService(BaseDjangoModelService['Entry']):
    obj: Entry

    @close_db_connections
    def convert_files_to_model_objects(self) -> str:
        file_objects = self.obj_class.services.tool.get_files_to_convert()

        entries = self.obj_class.objects.id_in(
            list({file_object.object_id for file_object in file_objects})
        ).select_related('current_version')

        entry_pk_map = {entry.pk: entry for entry in entries}

        errored = []
        for file_object in file_objects:
            try:
                EntryVersionBlock.services.factory.create_blocks_from_file(
                    file=file_object,
                    entry_version=entry_pk_map[file_object.object_id].current_version,
                )
            except Exception as e:
                errored.append({'file': file_object.name, 'error': traceback.format_exc()})
                file_object.set_deleted()
            else:
                file_object.set_deleted()

        message = f'Files Converted: {len(file_objects) - len(errored)}'

        if errored:
            error_string = f'\n{message}\nFiles Errored:'

            for error in errored:
                error_string += f'    File Name: {error["file"]}\n    Error: {error["error"]}'

            raise KnowledgeBaseConversionError(error_string)

        return message



from __future__ import annotations

from django_spire.auth.user.tests.factories import create_user
from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.choices import NotificationStatusChoices
from django_spire.notification.email.models import EmailNotification
from django_spire.notification.email.tests.factories import create_test_email_notification


class EmailNotificationQuerySetTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.user = create_user(username='test_email_queryset_user')
        self.email_notification = create_test_email_notification(
            user=self.user,
            status=NotificationStatusChoices.SENT
        )

    def test_by_user(self):
        other_user = create_user(username='other_user')
        other_notification = create_test_email_notification(user=other_user)

        result = EmailNotification.objects.by_user(self.user)
        assert self.email_notification in result
        assert other_notification not in result

    def test_by_users(self):
        other_user = create_user(username='other_user')
        other_notification = create_test_email_notification(user=other_user)

        result = EmailNotification.objects.by_users([self.user, other_user])
        assert self.email_notification in result
        assert other_notification in result

    def test_is_sent(self):
        pending_notification = create_test_email_notification(
            user=self.user,
            status=NotificationStatusChoices.PENDING
        )

        result = EmailNotification.objects.is_sent()
        assert self.email_notification in result
        assert pending_notification not in result



from django_spire.contrib.constructor.exceptions import ConstructorError
from django_spire.contrib.constructor.django_model_constructor import BaseDjangoModelConstructor
from django_spire.contrib.constructor.constructor import BaseConstructor

__all__ = [
    'BaseConstructor',
    'BaseDjangoModelConstructor',
    'ConstructorError'
]



from __future__ import annotations

from django_spire.ai.prompt.system import bots
from django_spire.ai.prompt.system.intel import SystemPromptIntel, SystemPromptResultIntel
from django_spire.core.tests.test_cases import BaseTestCase


class PromptBotTestCase(BaseTestCase):
    def test_role_system_prompt_bot(self) -> None:
        ISSUE_PROMPT = (
            'I am using jira to plan developer workloads And I want to be able to build very clear issues that '
            'developers will be able to read understand and take action on to complete the goal. '
            'I will give you details about a software development task that we have to achieve and I want you '
            'to return a short one sentence summary that the developer can read and understand the scope of the issue '
            'before they get into the details of it '
        )

        role_bot = bots.RoleSystemPromptBot()
        role_result = role_bot.process(ISSUE_PROMPT)

        assert role_result is not None
        assert role_result.result is not None
        assert len(role_result.result) > 0

    def test_task_system_prompt_bot(self) -> None:
        ISSUE_PROMPT = (
            'I am using jira to plan developer workloads And I want to be able to build very clear issues that '
            'developers will be able to read understand and take action on to complete the goal.'
        )

        task_bot = bots.TaskSystemPromptBot()
        task_result = task_bot.process(ISSUE_PROMPT)

        assert task_result is not None
        assert task_result.result is not None
        assert len(task_result.result) > 0

    def test_guidelines_system_prompt_bot(self) -> None:
        ISSUE_PROMPT = (
            'I want to create a system that helps users write better emails.'
        )

        guidelines_bot = bots.GuidelinesSystemPromptBot()
        guidelines_result = guidelines_bot.process(ISSUE_PROMPT)

        assert guidelines_result is not None
        assert guidelines_result.result is not None
        assert len(guidelines_result.result) > 0

    def test_output_format_system_prompt_bot(self) -> None:
        ISSUE_PROMPT = (
            'I want to create a system that returns a summary of a document.'
        )

        output_format_bot = bots.OutputFormatSystemPromptBot()
        output_format_result = output_format_bot.process(ISSUE_PROMPT)

        assert output_format_result is not None
        assert output_format_result.result is not None

    def test_system_prompt_bot_full_workflow(self) -> None:
        ISSUE_PROMPT = (
            'I am using jira to plan developer workloads And I want to be able to build very clear issues that '
            'developers will be able to read understand and take action on to complete the goal. '
            'I will give you details about a software development task that we have to achieve and I want you '
            'to return a short one sentence summary that the developer can read and understand the scope of the issue '
            'before they get into the details of it '
        )

        role_bot = bots.RoleSystemPromptBot()
        role_result = role_bot.process(ISSUE_PROMPT)
        assert role_result is not None

        task_bot = bots.TaskSystemPromptBot()
        task_result = task_bot.process(ISSUE_PROMPT)
        assert task_result is not None

        guidelines_bot = bots.GuidelinesSystemPromptBot()
        guidelines_result = guidelines_bot.process(ISSUE_PROMPT)
        assert guidelines_result is not None

        output_format_bot = bots.OutputFormatSystemPromptBot()
        output_format_result = output_format_bot.process(ISSUE_PROMPT)
        assert output_format_result is not None


class SystemPromptIntelTestCase(BaseTestCase):
    def test_system_prompt_intel_to_string(self) -> None:
        intel = SystemPromptIntel(
            role='Test Role',
            task='Test Task',
            guidelines='Test Guidelines',
            output_format='Test Output Format'
        )

        result = intel.to_string()

        assert 'Role: Test Role' in result
        assert 'Task: Test Task' in result
        assert 'Guidelines: Test Guidelines' in result
        assert 'Output Format: Test Output Format' in result

    def test_system_prompt_intel_to_string_without_output_format(self) -> None:
        intel = SystemPromptIntel(
            role='Test Role',
            task='Test Task',
            guidelines='Test Guidelines',
            output_format=None
        )

        result = intel.to_string()

        assert 'Role: Test Role' in result
        assert 'Task: Test Task' in result
        assert 'Guidelines: Test Guidelines' in result
        assert 'Output Format' not in result

    def test_system_prompt_result_intel(self) -> None:
        intel = SystemPromptResultIntel(result='Test result')

        assert intel.result == 'Test result'



from __future__ import annotations

from unittest.mock import patch

from django.urls import reverse

from django_spire.ai.sms.models import SmsConversation
from django_spire.core.tests.test_cases import BaseTestCase


class SmsWebhookTests(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.webhook_url = reverse('django_spire:ai:sms:webhook')

    @patch('twilio.request_validator.RequestValidator.validate')
    def test_webhook_receives_message(self, mock_validate) -> None:
        mock_validate.return_value = True

        response = self.client.post(
            self.webhook_url,
            {
                'From': '+15551234567',
                'Body': 'Hello',
                'MessageSid': 'SM123456789'
            }
        )

        assert response.status_code == 200

        conversation = SmsConversation.objects.get(phone_number='+15551234567')
        assert conversation.messages.count() == 2

        inbound_message = conversation.messages.filter(is_inbound=True).first()
        assert inbound_message.body == 'Hello'
        assert inbound_message.twilio_sid == 'SM123456789'

        outbound_message = conversation.messages.filter(is_inbound=False).first()
        assert len(outbound_message.body) > 0

    @patch('twilio.request_validator.RequestValidator.validate')
    def test_webhook_creates_conversation(self, mock_validate) -> None:
        mock_validate.return_value = True

        initial_count = SmsConversation.objects.count()

        self.client.post(
            self.webhook_url,
            {
                'From': '+15559999999',
                'Body': 'New conversation',
                'MessageSid': 'SM999999999'
            }
        )

        assert SmsConversation.objects.count() == initial_count + 1

    @patch('twilio.request_validator.RequestValidator.validate')
    def test_webhook_reuses_existing_conversation(self, mock_validate) -> None:
        mock_validate.return_value = True

        SmsConversation.objects.create(phone_number='+15551234567')

        initial_count = SmsConversation.objects.count()

        self.client.post(
            self.webhook_url,
            {
                'From': '+15551234567',
                'Body': 'Another message',
                'MessageSid': 'SM123456789'
            }
        )

        assert SmsConversation.objects.count() == initial_count

    @patch('twilio.request_validator.RequestValidator.validate')
    def test_webhook_invalid_signature(self, mock_validate) -> None:
        mock_validate.return_value = False

        response = self.client.post(
            self.webhook_url,
            {
                'From': '+15551234567',
                'Body': 'Hello',
                'MessageSid': 'SM123456789'
            }
        )

        assert response.status_code == 403

    @patch('twilio.request_validator.RequestValidator.validate')
    def test_webhook_short_phone_number(self, mock_validate) -> None:
        mock_validate.return_value = True

        response = self.client.post(
            self.webhook_url,
            {
                'From': '1234',
                'Body': 'Hello',
                'MessageSid': 'SM123456789'
            }
        )

        assert response.status_code == 403

    @patch('twilio.request_validator.RequestValidator.validate')
    def test_webhook_empty_body(self, mock_validate) -> None:
        mock_validate.return_value = True

        response = self.client.post(
            self.webhook_url,
            {
                'From': '+15551234567',
                'Body': '',
                'MessageSid': 'SM123456789'
            }
        )

        assert response.status_code == 200

    @patch('twilio.request_validator.RequestValidator.validate')
    def test_webhook_marks_inbound_as_processed(self, mock_validate) -> None:
        mock_validate.return_value = True

        self.client.post(
            self.webhook_url,
            {
                'From': '+15551234567',
                'Body': 'Test message',
                'MessageSid': 'SM123456789'
            }
        )

        conversation = SmsConversation.objects.get(phone_number='+15551234567')
        inbound_message = conversation.messages.filter(is_inbound=True).first()

        assert inbound_message.is_processed

    @patch('twilio.request_validator.RequestValidator.validate')
    def test_webhook_response_is_twiml(self, mock_validate) -> None:
        mock_validate.return_value = True

        response = self.client.post(
            self.webhook_url,
            {
                'From': '+15551234567',
                'Body': 'Hello',
                'MessageSid': 'SM123456789'
            }
        )

        assert b'<Response>' in response.content or b'Response' in response.content

    @patch('twilio.request_validator.RequestValidator.validate')
    def test_webhook_missing_from(self, mock_validate) -> None:
        mock_validate.return_value = True

        response = self.client.post(
            self.webhook_url,
            {
                'Body': 'Hello',
                'MessageSid': 'SM123456789'
            }
        )

        assert response.status_code == 403

    @patch('twilio.request_validator.RequestValidator.validate')
    def test_webhook_stores_twilio_sid(self, mock_validate) -> None:
        mock_validate.return_value = True

        self.client.post(
            self.webhook_url,
            {
                'From': '+15551234567',
                'Body': 'Hello',
                'MessageSid': 'SM_TEST_SID_123'
            }
        )

        conversation = SmsConversation.objects.get(phone_number='+15551234567')
        inbound_message = conversation.messages.filter(is_inbound=True).first()

        assert inbound_message.twilio_sid == 'SM_TEST_SID_123'



from __future__ import annotations

from unittest.mock import MagicMock, patch

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.help_desk.choices import (
    HelpDeskTicketPriorityChoices,
    HelpDeskTicketPurposeChoices,
    HelpDeskTicketStatusChoices,
)
from django_spire.help_desk.models import HelpDeskTicket
from django_spire.help_desk.tests.factories import create_test_helpdesk_ticket


class HelpDeskTicketServiceTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.ticket = create_test_helpdesk_ticket()

    def test_save_model_obj_updates_existing(self):
        self.ticket.services.obj = self.ticket
        updated_ticket, created = self.ticket.services.save_model_obj(
            description='Updated description'
        )

        assert created is False
        assert updated_ticket.description == 'Updated description'

    def test_save_model_obj_updates_priority(self):
        self.ticket.services.obj = self.ticket
        updated_ticket, _ = self.ticket.services.save_model_obj(
            priority=HelpDeskTicketPriorityChoices.URGENT
        )

        assert updated_ticket.priority == HelpDeskTicketPriorityChoices.URGENT

    def test_save_model_obj_updates_status(self):
        self.ticket.services.obj = self.ticket
        updated_ticket, _ = self.ticket.services.save_model_obj(
            status=HelpDeskTicketStatusChoices.DONE
        )

        assert updated_ticket.status == HelpDeskTicketStatusChoices.DONE

    @patch(
        'django_spire.help_desk.services.notification_service.'
        'HelpDeskTicketNotificationService.create_new_ticket_notifications'
    )
    def test_create_sets_created_by(self, _mock_notifications: MagicMock):
        ticket = HelpDeskTicket()

        created_ticket = ticket.services.create(
            created_by=self.super_user,
            purpose=HelpDeskTicketPurposeChoices.APP,
            priority=HelpDeskTicketPriorityChoices.LOW,
            description='Test ticket',
        )

        assert created_ticket.created_by == self.super_user

    @patch(
        'django_spire.help_desk.services.notification_service.'
        'HelpDeskTicketNotificationService.create_new_ticket_notifications'
    )
    def test_create_calls_notification_service(self, mock_notifications: MagicMock):
        ticket = HelpDeskTicket()

        ticket.services.create(
            created_by=self.super_user,
            purpose=HelpDeskTicketPurposeChoices.APP,
            priority=HelpDeskTicketPriorityChoices.LOW,
            description='Test ticket',
        )

        mock_notifications.assert_called_once()

    @patch(
        'django_spire.help_desk.services.notification_service.'
        'HelpDeskTicketNotificationService.create_new_ticket_notifications'
    )
    def test_create_returns_ticket(self, _mock_notifications: MagicMock):
        ticket = HelpDeskTicket()

        created_ticket = ticket.services.create(
            created_by=self.super_user,
            purpose=HelpDeskTicketPurposeChoices.APP,
            priority=HelpDeskTicketPriorityChoices.LOW,
            description='Test ticket',
        )

        assert created_ticket.pk is not None
        assert created_ticket.purpose == HelpDeskTicketPurposeChoices.APP



from __future__ import annotations

from django.db import models

from django_spire.file.models import File
from django_spire.notification.email.querysets import EmailNotificationQuerySet
from django_spire.notification.models import Notification


class EmailNotification(models.Model):
    """
    It is important to note size limits for email content contained in this model. E.g., Sendgrid has a hard total email
    limit of 30mb (and a recommended limit of 10mb for attachments): https://www.twilio.com/docs/sendgrid/ui/sending-email/attachments-with-digioh#-Limitations

    """

    notification = models.OneToOneField(
        Notification,
        editable=False,
        on_delete=models.CASCADE,
        related_name='email',
        related_query_name='email',
    )

    attachments = models.ManyToManyField(
        File,
        blank=True,
        related_name='attachments',
        related_query_name='attachment',
    )

    to_email_address = models.EmailField()
    template_id = models.CharField(max_length=64, default='')
    context_data = models.JSONField(default=dict)
    cc = models.JSONField(default=list)
    bcc = models.JSONField(default=list)

    objects = EmailNotificationQuerySet.as_manager()

    def __str__(self):
        return f'{self.notification.title} - {self.to_email_address}'

    class Meta:
        db_table = 'django_spire_notification_email'
        verbose_name = 'Email Notification'
        verbose_name_plural = 'Email Notifications'



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.sms.exceptions import (
    SmsNotificationError,
    SmsTemporaryMediaError,
    TwilioAPIConcurrentError,
    TwilioError,
)


class SmsExceptionsTests(BaseTestCase):
    def test_sms_notification_error_message(self):
        error = SmsNotificationError('SMS error')
        assert str(error) == 'SMS error'

    def test_sms_temporary_media_error_message(self):
        error = SmsTemporaryMediaError('Media error')
        assert str(error) == 'Media error'

    def test_twilio_error_message(self):
        error = TwilioError('Twilio error')
        assert str(error) == 'Twilio error'

    def test_twilio_api_concurrent_error_message(self):
        error = TwilioAPIConcurrentError('Concurrent error')
        assert str(error) == 'Concurrent error'



from __future__ import annotations

from copy import copy

from django.test import TestCase

from django_spire.contrib.seeding.field.cleaners import normalize_seeder_fields


FIELDS = {
    'faker_tuple': ('faker', 'name'),
    'llm_type': 'llm',
    'faker_type': 'faker',
    'static_type': 'static',
    'callable_type': 'callable',
    'custom_type': 'custom',
    'static_bool': True,
    'static_str': 'approved',
    'static_int': 10,
    'callable_func': lambda: 'now',
    'exclude_str': 'exclude',
    'exclude_tuple': ('exclude',)
}


class TestNormalizeSeederFields(TestCase):
    def test_excludes_fields(self) -> None:
        fields = copy(FIELDS)

        normalized = normalize_seeder_fields(fields)

        assert 'exclude_str' not in normalized
        assert 'exclude_tuple' not in normalized
        assert 'static_str' in normalized

    def test_normalizes_callable(self) -> None:
        fields = {'callable_func': FIELDS['callable_func']}

        normalized = normalize_seeder_fields(fields)

        assert normalized['callable_func'][0] == 'callable'
        assert callable(normalized['callable_func'][1])

    def test_normalizes_faker_tuple(self) -> None:
        fields = {'faker_tuple': FIELDS['faker_tuple']}

        normalized = normalize_seeder_fields(fields)

        assert normalized['faker_tuple'] == ('faker', 'name')

    def test_normalizes_single_value_strings_to_tuples(self) -> None:
        fields = {
            k: FIELDS[k] for k in (
                'llm_type', 'faker_type', 'static_type', 'callable_type', 'custom_type'
            )
        }

        normalized = normalize_seeder_fields(fields)

        for key, value in fields.items():
            assert normalized[key] == (value,)

    def test_normalizes_static_values(self) -> None:
        fields = {
            'static_bool': FIELDS['static_bool'],
            'static_str': FIELDS['static_str'],
            'static_int': FIELDS['static_int']
        }

        normalized = normalize_seeder_fields(fields)

        assert normalized['static_bool'] == ('static', True)
        assert normalized['static_str'] == ('static', 'approved')
        assert normalized['static_int'] == ('static', 10)



from __future__ import annotations

import pytest

from unittest.mock import MagicMock

from django.test import TestCase

from django_spire.contrib.seeding.model.config import FieldsConfig


class TestFieldsConfig(TestCase):
    def setUp(self) -> None:
        self.model_class = MagicMock()

    def test_assign_defaults_adds_missing_fields(self) -> None:
        config = FieldsConfig(
            raw_fields={'name': ('static', 'test')},
            field_names=['name', 'email'],
            default_to='llm',
            model_class=self.model_class
        )

        assert 'email' in config.fields
        assert config.fields['email'] == ('llm',)

    def test_excluded_property(self) -> None:
        config = FieldsConfig(
            raw_fields={'name': 'exclude', 'email': ('exclude',)},
            field_names=['name', 'email', 'status'],
            default_to='llm',
            model_class=self.model_class
        )

        assert 'name' in config.excluded
        assert 'email' in config.excluded
        assert 'status' not in config.excluded

    def test_fields_are_ordered(self) -> None:
        config = FieldsConfig(
            raw_fields={'zebra': ('static', 'z'), 'apple': ('static', 'a')},
            field_names=['zebra', 'apple'],
            default_to='included',
            model_class=self.model_class
        )

        keys = list(config.fields.keys())

        assert keys == ['apple', 'zebra']

    def test_override_creates_new_config(self) -> None:
        config = FieldsConfig(
            raw_fields={'name': ('static', 'original')},
            field_names=['name', 'email'],
            default_to='llm',
            model_class=self.model_class
        )

        new_config = config.override({'name': ('static', 'overridden')})

        assert new_config is not config
        assert new_config.fields['name'] == ('static', 'overridden')

    def test_validate_raises_for_invalid_fields(self) -> None:
        with pytest.raises(ValueError, match='Invalid field name'):
            FieldsConfig(
                raw_fields={'invalid_field': ('static', 'test')},
                field_names=['name', 'email'],
                default_to='llm',
                model_class=self.model_class
            )



from __future__ import annotations

import pytest

from django_spire.auth.group.models import AuthGroup
from django_spire.auth.group.utils import (
    codename_list_to_perm_level,
    codename_to_perm_level,
    has_app_permission,
    has_app_permission_or_404,
    perm_level_to_django_permission,
    perm_level_to_int,
    perm_level_to_string,
    set_group_users,
)
from django_spire.auth.user.tests.factories import create_super_user, create_user
from django_spire.core.tests.test_cases import BaseTestCase


class SetGroupUsersTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.group = AuthGroup.objects.create(name='Test Group')
        self.user1 = create_user(username='user1')
        self.user2 = create_user(username='user2')
        self.user3 = create_user(username='user3')

    def test_set_group_users(self) -> None:
        set_group_users(self.group, [self.user1, self.user2])
        assert self.group.user_set.count() == 2

    def test_set_group_users_empty(self) -> None:
        self.group.user_set.add(self.user1)
        set_group_users(self.group, [])
        assert self.group.user_set.count() == 0

    def test_set_group_users_replaces_existing(self) -> None:
        self.group.user_set.add(self.user1)
        set_group_users(self.group, [self.user2])
        assert self.group.user_set.count() == 1
        assert self.user2 in self.group.user_set.all()
        assert self.user1 not in self.group.user_set.all()

    def test_set_group_users_multiple(self) -> None:
        set_group_users(self.group, [self.user1, self.user2, self.user3])
        assert self.group.user_set.count() == 3

    def test_set_group_users_idempotent(self) -> None:
        set_group_users(self.group, [self.user1])
        set_group_users(self.group, [self.user1])
        assert self.group.user_set.count() == 1

    def test_set_group_users_preserves_user_in_other_groups(self) -> None:
        other_group = AuthGroup.objects.create(name='Other Group')
        other_group.user_set.add(self.user1)
        set_group_users(self.group, [self.user1])
        assert self.user1 in other_group.user_set.all()
        assert self.user1 in self.group.user_set.all()

    def test_set_group_users_with_queryset(self) -> None:
        from django_spire.auth.user.models import AuthUser
        users = AuthUser.objects.filter(pk__in=[self.user1.pk, self.user2.pk])
        set_group_users(self.group, list(users))
        assert self.group.user_set.count() == 2

    def test_set_group_users_clears_all_then_sets(self) -> None:
        self.group.user_set.add(self.user1, self.user2, self.user3)
        set_group_users(self.group, [self.user1])
        assert self.group.user_set.count() == 1
        assert self.user1 in self.group.user_set.all()


class CodenameToPermLevelTestCase(BaseTestCase):
    def test_view_codename(self) -> None:
        assert codename_to_perm_level('view_model') == 1

    def test_add_codename(self) -> None:
        assert codename_to_perm_level('add_model') == 2

    def test_change_codename(self) -> None:
        assert codename_to_perm_level('change_model') == 3

    def test_delete_codename(self) -> None:
        assert codename_to_perm_level('delete_model') == 4

    def test_none_codename(self) -> None:
        assert codename_to_perm_level('none_model') == 0

    def test_unknown_prefix(self) -> None:
        assert codename_to_perm_level('unknown_model') == 0

    def test_complex_model_name(self) -> None:
        assert codename_to_perm_level('view_my_complex_model') == 1
        assert codename_to_perm_level('delete_my_complex_model') == 4

    def test_uppercase_prefix(self) -> None:
        assert codename_to_perm_level('VIEW_model') == 1

    def test_mixed_case_prefix(self) -> None:
        assert codename_to_perm_level('View_model') == 1

    def test_empty_codename(self) -> None:
        assert codename_to_perm_level('') == 0

    def test_underscore_only(self) -> None:
        assert codename_to_perm_level('_') == 0

    def test_no_underscore(self) -> None:
        assert codename_to_perm_level('viewmodel') == 0

    def test_multiple_underscores(self) -> None:
        assert codename_to_perm_level('view_my_model_name') == 1

    def test_special_characters(self) -> None:
        assert codename_to_perm_level('view_model@special') == 1


class CodenameListToPermLevelTestCase(BaseTestCase):
    def test_empty_list(self) -> None:
        assert codename_list_to_perm_level([]) == 0

    def test_single_codename(self) -> None:
        assert codename_list_to_perm_level(['view_model']) == 1

    def test_multiple_codenames_returns_highest(self) -> None:
        codenames = ['view_model', 'add_model', 'change_model']
        assert codename_list_to_perm_level(codenames) == 3

    def test_all_permissions(self) -> None:
        codenames = ['view_model', 'add_model', 'change_model', 'delete_model']
        assert codename_list_to_perm_level(codenames) == 4

    def test_unordered_codenames(self) -> None:
        codenames = ['delete_model', 'view_model', 'add_model']
        assert codename_list_to_perm_level(codenames) == 4

    def test_duplicate_codenames(self) -> None:
        codenames = ['view_model', 'view_model', 'view_model']
        assert codename_list_to_perm_level(codenames) == 1

    def test_mixed_valid_invalid(self) -> None:
        codenames = ['view_model', 'invalid_model', 'delete_model']
        assert codename_list_to_perm_level(codenames) == 4

    def test_all_invalid(self) -> None:
        codenames = ['invalid_model', 'unknown_model', 'fake_model']
        assert codename_list_to_perm_level(codenames) == 0

    def test_single_delete(self) -> None:
        assert codename_list_to_perm_level(['delete_model']) == 4

    def test_view_and_add(self) -> None:
        codenames = ['view_model', 'add_model']
        assert codename_list_to_perm_level(codenames) == 2


class PermLevelToIntTestCase(BaseTestCase):
    def test_integer_zero(self) -> None:
        assert perm_level_to_int(0) == 0

    def test_integer_valid_range(self) -> None:
        for i in range(5):
            assert perm_level_to_int(i) == i

    def test_integer_out_of_range_high(self) -> None:
        assert perm_level_to_int(5) == 0
        assert perm_level_to_int(100) == 0

    def test_integer_out_of_range_low(self) -> None:
        assert perm_level_to_int(-1) == 0
        assert perm_level_to_int(-100) == 0

    def test_string_none(self) -> None:
        assert perm_level_to_int('none') == 0
        assert perm_level_to_int('None') == 0
        assert perm_level_to_int('NONE') == 0

    def test_string_view(self) -> None:
        assert perm_level_to_int('view') == 1
        assert perm_level_to_int('View') == 1
        assert perm_level_to_int('VIEW') == 1

    def test_string_add(self) -> None:
        assert perm_level_to_int('add') == 2
        assert perm_level_to_int('Add') == 2
        assert perm_level_to_int('ADD') == 2

    def test_string_change(self) -> None:
        assert perm_level_to_int('change') == 3
        assert perm_level_to_int('Change') == 3
        assert perm_level_to_int('CHANGE') == 3

    def test_string_delete(self) -> None:
        assert perm_level_to_int('delete') == 4
        assert perm_level_to_int('Delete') == 4
        assert perm_level_to_int('DELETE') == 4

    def test_invalid_string(self) -> None:
        assert perm_level_to_int('invalid') == 0
        assert perm_level_to_int('') == 0
        assert perm_level_to_int('random') == 0

    def test_whitespace_string(self) -> None:
        assert perm_level_to_int('  view  ') == 0

    def test_string_with_numbers(self) -> None:
        assert perm_level_to_int('view1') == 0


class PermLevelToStringTestCase(BaseTestCase):
    def test_level_zero(self) -> None:
        assert perm_level_to_string(0) == 'None'

    def test_level_one(self) -> None:
        assert perm_level_to_string(1) == 'View'

    def test_level_two(self) -> None:
        assert perm_level_to_string(2) == 'Add'

    def test_level_three(self) -> None:
        assert perm_level_to_string(3) == 'Change'

    def test_level_four(self) -> None:
        assert perm_level_to_string(4) == 'Delete'

    def test_string_input(self) -> None:
        assert perm_level_to_string('view') == 'View'
        assert perm_level_to_string('delete') == 'Delete'

    def test_invalid_level_returns_none(self) -> None:
        assert perm_level_to_string(5) == 'None'
        assert perm_level_to_string(-1) == 'None'

    def test_all_string_inputs(self) -> None:
        assert perm_level_to_string('none') == 'None'
        assert perm_level_to_string('view') == 'View'
        assert perm_level_to_string('add') == 'Add'
        assert perm_level_to_string('change') == 'Change'
        assert perm_level_to_string('delete') == 'Delete'

    def test_very_high_number(self) -> None:
        assert perm_level_to_string(999) == 'None'

    def test_very_low_number(self) -> None:
        assert perm_level_to_string(-999) == 'None'


class PermLevelToDjangoPermissionTestCase(BaseTestCase):
    def test_view_permission(self) -> None:
        result = perm_level_to_django_permission(1, 'app_label', 'model')
        assert result == 'app_label.view_model'

    def test_add_permission(self) -> None:
        result = perm_level_to_django_permission(2, 'app_label', 'model')
        assert result == 'app_label.add_model'

    def test_change_permission(self) -> None:
        result = perm_level_to_django_permission(3, 'app_label', 'model')
        assert result == 'app_label.change_model'

    def test_delete_permission(self) -> None:
        result = perm_level_to_django_permission(4, 'app_label', 'model')
        assert result == 'app_label.delete_model'

    def test_none_permission(self) -> None:
        result = perm_level_to_django_permission(0, 'app_label', 'model')
        assert result == 'app_label.none_model'

    def test_complex_app_label(self) -> None:
        result = perm_level_to_django_permission(1, 'my_app_label', 'model')
        assert result == 'my_app_label.view_model'

    def test_complex_model_name(self) -> None:
        result = perm_level_to_django_permission(1, 'app', 'my_model_name')
        assert result == 'app.view_my_model_name'

    def test_string_perm_level(self) -> None:
        result = perm_level_to_django_permission('view', 'app', 'model')
        assert result == 'app.view_model'

    def test_uppercase_string_perm_level(self) -> None:
        result = perm_level_to_django_permission('DELETE', 'app', 'model')
        assert result == 'app.delete_model'


class HasAppPermissionTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.user = create_user(username='testuser')

    def test_user_without_permission(self) -> None:
        result = has_app_permission(self.user, 'app_label', 'model', 'view')
        assert result is False

    def test_superuser_has_permission(self) -> None:
        superuser = create_super_user()
        result = has_app_permission(superuser, 'any_app', 'any_model', 'delete')
        assert result is True

    def test_various_actions(self) -> None:
        for action in ['view', 'add', 'change', 'delete']:
            result = has_app_permission(self.user, 'app', 'model', action)
            assert result is False

    def test_superuser_all_actions(self) -> None:
        superuser = create_super_user()
        for action in ['view', 'add', 'change', 'delete']:
            result = has_app_permission(superuser, 'app', 'model', action)
            assert result is True

    def test_invalid_action(self) -> None:
        result = has_app_permission(self.user, 'app', 'model', 'invalid')
        assert result is False


class HasAppPermissionOr404TestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.user = create_user(username='testuser')

    def test_raises_permission_error(self) -> None:
        with pytest.raises(PermissionError):
            has_app_permission_or_404(self.user, 'app', 'model', 'view')

    def test_superuser_returns_true(self) -> None:
        superuser = create_super_user()
        result = has_app_permission_or_404(superuser, 'any_app', 'any_model', 'delete')
        assert result is True

    def test_raises_for_all_actions(self) -> None:
        for action in ['view', 'add', 'change', 'delete']:
            with pytest.raises(PermissionError):
                has_app_permission_or_404(self.user, 'app', 'model', action)

    def test_superuser_all_actions_pass(self) -> None:
        superuser = create_super_user()
        for action in ['view', 'add', 'change', 'delete']:
            result = has_app_permission_or_404(superuser, 'app', 'model', action)
            assert result is True



from __future__ import annotations

from django.urls import path

from django_spire.notification.app.views import json_views


app_name = 'django_spire_notification'

urlpatterns = [
    path('check/notification/ajax/',
        json_views.check_new_notifications_ajax_view,
        name='check_new'),

    path('set_viewed/notification/ajax/',
        json_views.set_notifications_as_viewed_ajax,
        name='set_viewed'),
]



from __future__ import annotations

from django.contrib.contenttypes.models import ContentType

from django_spire.comment.forms import CommentForm
from django_spire.core.tests.test_cases import BaseTestCase


class TestCommentForm(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.content_type = ContentType.objects.get_for_model(self.super_user)

    def test_valid_form(self) -> None:
        form = CommentForm(data={'information': 'Test comment'})
        assert form.is_valid() is True

    def test_form_save(self) -> None:
        form = CommentForm(data={'information': 'New comment'})
        assert form.is_valid() is True
        comment = form.save(commit=False)
        comment.content_type = self.content_type
        comment.object_id = self.super_user.pk
        comment.user = self.super_user
        comment.save()
        assert comment.information == 'New comment'



from __future__ import annotations

from django.test import TestCase, override_settings

from django_spire.conf import Settings, settings


class TestSettings(TestCase):
    def test_getattr_returns_django_setting(self) -> None:
        assert settings.DEBUG is not None

    def test_getattr_returns_django_spire_default(self) -> None:
        result = settings.DJANGO_SPIRE_DEFAULT_THEME

        assert result == 'default-light'

    def test_getattr_returns_none_for_unknown_setting(self) -> None:
        result = settings.NONEXISTENT_SETTING_NAME

        assert result is None

    @override_settings(DJANGO_SPIRE_DEFAULT_THEME='custom-dark')
    def test_django_setting_overrides_default(self) -> None:
        fresh_settings = Settings()

        result = fresh_settings.DJANGO_SPIRE_DEFAULT_THEME

        assert result == 'custom-dark'

    @override_settings(
        DJANGO_SPIRE_AUTH_CONTROLLERS={'custom_app': 'custom.path.Controller'}
    )
    def test_auth_controllers_merges_dicts(self) -> None:
        fresh_settings = Settings()

        result = fresh_settings.DJANGO_SPIRE_AUTH_CONTROLLERS

        assert 'custom_app' in result
        assert 'ai_chat' in result

    def test_settings_instance_exists(self) -> None:
        assert settings is not None
        assert isinstance(settings, Settings)



from typing import Type, Tuple

from django.contrib import admin
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models


class SpireModelAdmin(admin.ModelAdmin):
    model_class: Type[models.Model] = None

    max_search_fields: int = 5
    max_list_display: int = 10

    trailing_fields = ('is_active', 'is_deleted')

    auto_readonly_fields: Tuple[str] = (
        'created_datetime', 'is_active', 'is_deleted',
    )

    filter_field_types = (
        models.BooleanField,
        models.DateField,
        models.DateTimeField,
        models.ForeignKey,
        models.CharField,
    )

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        cls.model_fields = cls.model_class._meta.get_fields()

        if cls.model_class is None:
            raise ValueError(f'{cls.__name__} must define model_class')

        if cls.model_class is not None:
            cls._configure_if_needed()

    @classmethod
    def _configure_if_needed(cls):
        if not hasattr(cls, '_spire_configured'):
            cls._configure_list_display()
            cls._configure_list_filter()
            cls._configure_search_fields()
            cls._configure_readonly_fields()
            cls._configure_ordering()
            cls._configure_list_per_page()
            cls._spire_configured = True

    @classmethod
    def _configure_list_display(cls):
        if cls.list_display != ('__str__',):
            return

        fields = []

        for field in cls.model_fields:
            if not isinstance(
                field,
                (models.ManyToManyField, models.ManyToOneRel, GenericRelation),
            ):
                if hasattr(field, 'name') and not field.name.startswith('_'):
                    if field.name not in cls.trailing_fields:
                        fields.append(field.name)

        for trailing_field in cls.trailing_fields:
            if trailing_field in [field.name for field in cls.model_fields]:
                fields.append(trailing_field)

        cls.list_display = tuple(fields[:cls.max_list_display])

    @classmethod
    def _configure_list_filter(cls):
        if hasattr(cls, 'list_filter') and cls.list_filter:
            return

        filters = []

        for field in cls.model_fields:
            if not isinstance(field, (models.ManyToManyField, models.ManyToOneRel)):
                if isinstance(field, models.BooleanField):
                    filters.append(field.name)

                elif isinstance(field, (models.DateField, models.DateTimeField)):
                    filters.append(field.name)

                elif isinstance(field, models.ForeignKey):
                    filters.append(field.name)

                elif (
                    isinstance(field, models.CharField)
                    and hasattr(field, 'choices')
                    and field.choices
                ):
                    filters.append(field.name)

        cls.list_filter = tuple(filters)

    @classmethod
    def _configure_search_fields(cls):
        if hasattr(cls, 'search_fields') and cls.search_fields:
            return

        search_fields = []

        for field in cls.model_fields:
            if isinstance(field, (models.CharField, models.TextField)):
                if not field.name.startswith('_'):
                    search_fields.append(field.name)

                    if len(search_fields) >= cls.max_search_fields:
                        break

        cls.search_fields = tuple(search_fields)

    @classmethod
    def _configure_readonly_fields(cls):
        if hasattr(cls, 'readonly_fields') and cls.readonly_fields:
            return

        readonly = []

        for field in cls.model_fields:
            if hasattr(field, 'name') and field.name in cls.auto_readonly_fields:
                readonly.append(field.name)

        cls.readonly_fields = tuple(readonly)

    @classmethod
    def _configure_ordering(cls):
        if hasattr(cls, 'ordering') and cls.ordering:
            return

        cls.ordering = ('-id',)
        return

    @classmethod
    def _configure_list_per_page(cls):
        if not hasattr(cls, 'list_per_page') or cls.list_per_page == 100:
            cls.list_per_page = 25



from __future__ import annotations

from django.db import models


class HistoryEventChoices(models.TextChoices):
    CREATED = 'crea', 'Created'
    UPDATED = 'upda', 'Updated'
    ACTIVE = 'acti', 'Active'
    INACTIVE = 'inac', 'Inactive'
    DELETED = 'dele', 'Deleted'
    UNDELETED = 'unde', 'Un-Deleted'



from __future__ import annotations

from django.urls import include, path


app_name = 'sms'

urlpatterns = [
    path('media/', include('django_spire.notification.sms.urls.media_urls', namespace='media')),
]



from __future__ import annotations

from django.test import RequestFactory

from django_spire.auth.user.tests.factories import create_user
from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.collection.models import Collection
from django_spire.knowledge.collection.tests.factories import (
    create_test_auth_group,
    create_test_collection,
    create_test_collection_group,
)


class CollectionQuerySetTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()
        self.user = create_user(username='test_queryset_user')
        self.parent_collection = create_test_collection(name='Parent')
        self.child_collection = create_test_collection(
            parent=self.parent_collection,
            name='Child'
        )

    def test_by_parent(self):
        result = Collection.objects.by_parent(parent=self.parent_collection)
        assert self.child_collection in result
        assert self.parent_collection not in result

    def test_by_parent_id(self):
        result = Collection.objects.by_parent_id(parent_id=self.parent_collection.id)
        assert self.child_collection in result

    def test_by_parent_ids(self):
        result = Collection.objects.by_parent_ids(parent_ids=[self.parent_collection.id])
        assert self.child_collection in result

    def test_parentless(self):
        result = Collection.objects.parentless()
        assert self.parent_collection in result
        assert self.child_collection not in result

    def test_childless(self):
        grandchild = create_test_collection(
            parent=self.child_collection,
            name='Grandchild'
        )
        result = Collection.objects.childless()
        assert grandchild in result
        assert self.parent_collection not in result

    def test_children(self):
        grandchild = create_test_collection(
            parent=self.child_collection,
            name='Grandchild'
        )
        result = Collection.objects.children(collection_id=self.parent_collection.id)
        assert self.child_collection in result
        assert grandchild in result

    def test_exclude_children(self):
        grandchild = create_test_collection(
            parent=self.child_collection,
            name='Grandchild'
        )
        result = Collection.objects.exclude_children(collection_id=self.parent_collection.id)
        assert self.parent_collection in result
        assert self.child_collection not in result
        assert grandchild not in result

    def test_annotate_entry_count(self):
        result = Collection.objects.annotate_entry_count()
        assert result.first().entry_count == 0

    def test_request_user_has_access_superuser(self):
        request = self.factory.get('/')
        request.user = self.super_user
        result = Collection.objects.request_user_has_access(request)
        assert self.parent_collection in result

    def test_request_user_has_access_with_group(self):
        auth_group = create_test_auth_group(name='Access Group')
        self.user.groups.add(auth_group)
        create_test_collection_group(
            collection=self.parent_collection,
            auth_group=auth_group
        )

        request = self.factory.get('/')
        request.user = self.user
        result = Collection.objects.request_user_has_access(request)
        assert self.parent_collection in result



from __future__ import annotations

import uuid

from django.utils.timezone import now, timedelta

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.sms.choices import SmsMediaContentTypeChoices
from django_spire.notification.sms.models import SmsTemporaryMedia


class SmsTemporaryMediaTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.expired_media = SmsTemporaryMedia.objects.create(
            content='base64content',
            content_type=SmsMediaContentTypeChoices.PNG,
            name='expired.png',
            expire_datetime=now() - timedelta(hours=1),
            external_access_key=uuid.uuid4(),
        )
        self.valid_media = SmsTemporaryMedia.objects.create(
            content='base64content',
            content_type=SmsMediaContentTypeChoices.PNG,
            name='valid.png',
            expire_datetime=now() + timedelta(hours=1),
            external_access_key=uuid.uuid4(),
        )

    def test_is_expired_true(self):
        assert self.expired_media.is_expired() is True

    def test_is_expired_false(self):
        assert self.valid_media.is_expired() is False

    def test_external_url_contains_access_key(self):
        url = self.valid_media.external_url
        assert str(self.valid_media.external_access_key) in url



from __future__ import annotations

import json

from http import HTTPStatus
from unittest.mock import patch

from django.test import TestCase, RequestFactory
from django.http import JsonResponse

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.theme.enums import ThemeFamily
from django_spire.theme.utils import get_theme_cookie_name
from django_spire.theme.views import json_views


class ThemeViewTests(TestCase):
    def setUp(self) -> None:
        self.factory = RequestFactory()

    def test_get_config_success(self) -> None:
        request = self.factory.get('/django_spire/theme/json/get_config/')
        response = json_views.get_config(request)

        assert isinstance(response, JsonResponse)
        assert response.status_code == HTTPStatus.OK

        data = json.loads(response.content)
        assert data['success']

        config = data['data']
        assert 'families' in config
        assert 'default_family' in config
        assert 'default_mode' in config
        assert 'separator' in config

        for family in ThemeFamily:
            assert family.value in config['families']
            family_config = config['families'][family.value]
            assert 'name' in family_config
            assert 'modes' in family_config
            assert set(family_config['modes']) == {'dark', 'light'}

    def test_get_config_cache_header(self) -> None:
        _ = self.factory.get('/django_spire/theme/json/get_config/')

        with patch('django_spire.theme.views.json_views.cache_page') as _:
            assert hasattr(json_views.get_config, '__wrapped__')

    def test_set_theme_success(self) -> None:
        request = self.factory.post(
            '/django_spire/theme/json/set_theme/',
            data=json.dumps({'theme': 'gruvbox-dark'}),
            content_type='application/json'
        )
        response = json_views.set_theme(request)

        assert isinstance(response, JsonResponse)
        assert response.status_code == HTTPStatus.OK

        data = json.loads(response.content)
        assert data['success']
        assert 'theme' in data

        data = data['theme']
        assert data['family'] == 'gruvbox'
        assert data['mode'] == 'dark'

        cookie_name = get_theme_cookie_name()
        assert cookie_name in response.cookies
        assert response.cookies[cookie_name].value == 'gruvbox-dark'

    def test_set_theme_missing_theme(self) -> None:
        request = self.factory.post(
            '/django_spire/theme/json/set_theme/',
            data=json.dumps({}),
            content_type='application/json'
        )
        response = json_views.set_theme(request)

        assert response.status_code == HTTPStatus.BAD_REQUEST
        data = json.loads(response.content)
        assert not data['success']
        assert 'error' in data
        assert data['error'] == 'Theme is required'

    def test_set_theme_invalid_theme(self) -> None:
        request = self.factory.post(
            '/django_spire/theme/json/set_theme/',
            data=json.dumps({'theme': 'invalid-theme'}),
            content_type='application/json'
        )
        response = json_views.set_theme(request)

        assert response.status_code == HTTPStatus.BAD_REQUEST
        data = json.loads(response.content)
        assert not data['success']
        assert 'error' in data
        assert 'Invalid theme' in data['error']


class ThemeViewIntegrationTests(BaseTestCase):
    def test_set_theme_with_authenticated_client(self) -> None:
        response = self.client.post(
            '/django_spire/theme/json/set_theme/',
            data=json.dumps({'theme': 'gruvbox-dark'}),
            content_type='application/json'
        )

        assert response.status_code == HTTPStatus.OK
        data = json.loads(response.content)
        assert data['success']

    def test_get_config_with_authenticated_client(self) -> None:
        response = self.client.get('/django_spire/theme/json/get_config/')
        assert response.status_code == HTTPStatus.OK

        data = json.loads(response.content)
        assert data['success']
        assert 'families' in data['data']



from __future__ import annotations

from django_spire.auth.group.models import AuthGroup
from django_spire.knowledge.collection.models import Collection, CollectionGroup


def create_test_auth_group(**kwargs) -> AuthGroup:
    data = {
        'name': 'Test Group',
    }
    data.update(kwargs)
    return AuthGroup.objects.create(**data)


def create_test_collection(**kwargs) -> Collection:
    data = {
        'parent': None,
        'name': 'Video Game Cheat Codes',
        'description': 'A collection of video game cheat codes.',
        'is_deleted': False,
        'is_active': True
    }
    data.update(kwargs)
    return Collection.objects.create(**data)


def create_test_collection_group(
    collection: Collection | None = None,
    auth_group: AuthGroup | None = None,
    **kwargs
) -> CollectionGroup:
    if collection is None:
        collection = create_test_collection()
    if auth_group is None:
        auth_group = create_test_auth_group()

    data = {
        'collection': collection,
        'auth_group': auth_group,
    }
    data.update(kwargs)
    return CollectionGroup.objects.create(**data)



from __future__ import annotations

from django.urls import path

from django_spire.knowledge.views import page_views


app_name = 'page'

urlpatterns = [
    path('', page_views.home_view, name='home'),
]



from __future__ import annotations

from django.apps import AppConfig
from django.conf import settings

from django_spire.notification.sms.consts import TWILIO_SMS_BATCH_SIZE_NAME
from django_spire.utils import check_required_apps


class NotificationSmsConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'django_spire.notification.sms'
    label = 'django_spire_notification_sms'

    REQUIRED_APPS = ('django_spire_core', 'django_spire_notification')

    def ready(self) -> None:
        if not isinstance(getattr(settings, TWILIO_SMS_BATCH_SIZE_NAME), int):
            message = f'"{TWILIO_SMS_BATCH_SIZE_NAME}" must be set in the django settings when using "{self.label}".'
            raise TypeError(message)

        check_required_apps(self.label)



from __future__ import annotations

from django.urls import path

from django_spire.auth.views import redirect_views


app_name = 'redirect'

urlpatterns = [
    path('login/',
         redirect_views.login_redirect_view,
         name='login'),

    path('logout/',
         redirect_views.logout_redirect_view,
         name='logout'),
]



from __future__ import annotations

import functools
import os

from typing import TYPE_CHECKING, Callable

from django.http import HttpResponseForbidden
from twilio.request_validator import RequestValidator

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from django.http import HttpResponse


def twilio_auth_required(func: Callable[..., HttpResponse]) -> Callable[..., HttpResponse]:
    @functools.wraps(func)
    def decorated_function(request: WSGIRequest, *args, **kwargs) -> HttpResponse:
        request_validator = RequestValidator(os.environ.get('TWILIO_AUTH_TOKEN', ''))

        absolute_uri = request.build_absolute_uri()

        if absolute_uri[:5] == 'http:':
            absolute_uri = 'https' + absolute_uri[4:]

        request_valid = request_validator.validate(
            absolute_uri,
            request.POST,
            request.META.get('HTTP_X_TWILIO_SIGNATURE', '')
        )

        if request_valid:
            return func(request, *args, **kwargs)

        return HttpResponseForbidden()

    return decorated_function



from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING

from dandy import Bot, Prompt

from django_spire.knowledge.intelligence.intel.answer_intel import AnswerIntel

if TYPE_CHECKING:
    from dandy.llm.request.message import MessageHistory

    from django_spire.knowledge.entry.models import Entry


class BlockType(StrEnum):
    HEADING = 'HEADING'
    SUBHEADING = 'SUBHEADING'


class MarkerType(StrEnum):
    ARTICLE = 'ARTICLE'
    END_ARTICLE = 'END ARTICLE'


def format_marker(marker: MarkerType, label: str | None = None) -> str:
    if label:
        return f'[{marker}: {label}]'

    return f'[{marker}]'


class KnowledgeAnswerBot(Bot):
    llm_role = 'Knowledge Entry Search Assistant'
    llm_task = 'Read through the knowledge and answer the users request.'
    llm_guidelines = (
        Prompt()
        .list([
            'Make sure the answer is relevant and reflects knowledge entries.',
            'Do not make up information use the provided knowledge entries as a source of truth.',
            'Use line breaks to separate sections of the answer and use 2 if you need to separate the section from the previous.',
            'Use the conversation history to understand context and references like "before that", "my last query", etc.',
            'When a user asks about an article or section by title, summarize the content of that article or section.',
            f'Content under a [{BlockType.HEADING}] or [{BlockType.SUBHEADING}] belongs to that section.',
        ])
    )
    llm_intel_class = AnswerIntel

    def process(
        self,
        user_input: str,
        entries: list[Entry],
        message_history: MessageHistory | None = None
    ) -> AnswerIntel:
        entry_prompt = Prompt()
        entry_prompt.sub_heading('User Request')
        entry_prompt.line_break()
        entry_prompt.text(f'{user_input}')
        entry_prompt.line_break()
        entry_prompt.sub_heading('Knowledge')
        entry_prompt.line_break()

        for entry in entries:
            entry_prompt.text(format_marker(MarkerType.ARTICLE, entry.name))

            for version_block in entry.current_version.blocks.all():
                text = version_block.render_to_text()

                if text and text != '\n':
                    entry_prompt.text(f'[{version_block.type.upper()}] {text}')

            entry_prompt.text(format_marker(MarkerType.END_ARTICLE))
            entry_prompt.line_break()

        return self.llm.prompt_to_intel(
            prompt=entry_prompt,
            message_history=message_history,
        )



from __future__ import annotations

import logging
from abc import ABC
from itertools import chain
from typing import Generic, TypeVar

from django.contrib.auth.models import User
from django.db import transaction
from django.db.models import Model, AutoField, FileField

from django_spire.contrib.constructor.django_model_constructor import \
    BaseDjangoModelConstructor
from django_spire.contrib.service.exceptions import ServiceError


log = logging.getLogger(__name__)

TypeDjangoModel_co = TypeVar('TypeDjangoModel_co', bound=Model, covariant=True)


class BaseDjangoModelService(
    BaseDjangoModelConstructor[TypeDjangoModel_co],
    ABC,
    Generic[TypeDjangoModel_co]
):
    def _set_non_m2m_fields(self, **field_data: dict):
        model_fields = self.obj._meta.fields

        file_field_list = []
        updated_fields = []

        for field in model_fields:
            if isinstance(field, AutoField) or field.name not in field_data:
                continue

            # Defer saving file-type fields until after the other fields, so a
            # callable upload_to can use the values from other fields (from django's construct_instance).
            if isinstance(field, FileField):
                file_field_list.append(field)
                updated_fields.append(field.name)
            else:
                field.save_form_data(self.obj, field_data[field.name])
                updated_fields.append(field.name)

        # Update foreign key id aliases in field_data for
        # related fields that weren't already updated above
        foreign_key_id_aliases = [
            f"{field.name}_id" for field in model_fields
            if f"{field.name}_id" in field_data and field.many_to_one and field.name not in updated_fields
        ]

        for field_name in foreign_key_id_aliases:
            setattr(self.obj, field_name, field_data[field_name])

        # Update file fields deferred from earlier
        for field in file_field_list:
            field.save_form_data(self.obj, field_data[field.name])

    def _set_m2m_fields(self, **field_data):
        model_meta = self.obj._meta

        for field in chain(model_meta.many_to_many, model_meta.private_fields):
            if not hasattr(field, "save_form_data"):
                continue
            if field.name in field_data:
                field.save_form_data(self.obj, field_data[field.name])

    @transaction.atomic
    def save_model_obj(self, **field_data: dict | None) -> tuple[Model, bool]:
        """
        This is the core service method for saving a Django model object with field data provided via kwargs.
        It will update the object with the given kwargs and handle any upstream attribute changes that were applied
        directly to the model instance (i.e. it can also be called without any kwargs, similar to `Model.save()`).

        Its purpose is to run extra operations related to the model instance that need to run each time
        the model is saved - it is meant to replace the need for overriding `BaseModelForm.save()` or `Model.save()`.

        It is designed to emulate django's `BaseModelForm.save()` method as close as possible:
        first, in `_set_non_m2m_fields`, it updates the fields on `self.obj` using logic similar to `django.forms.models.construct_instance`,
        then it calls self.obj.save(), then updates the m2m fields on the object instance
        using logic similar to django's `BaseModelForm._save_m2m()` method. In all cases,
        it treats the incoming `field_data` exactly the same as `cleaned_data` is treated
        in the django code that it is emulating, and therefore does not perform any validation
        on the data or the model instance as it is assumed that field_data has already been validated upstream.

        Args:
            **field_data:

        Returns:
            tuple[Model, bool]

        """

        new_model_obj_was_created = True if self.model_obj_is_new else False

        self._set_non_m2m_fields(**field_data)
        self.obj.save()
        self._set_m2m_fields(**field_data)

        return self.obj, new_model_obj_was_created



from __future__ import annotations

from django.urls import path

from django_spire.knowledge.collection.views import json_views


app_name = 'json'

urlpatterns = [
    path('reorder/', json_views.reorder_view, name='reorder'),
]



# Generated by Django 5.2 on 2025-05-06 21:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_spire_ai_chat', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='chatmessage',
            name='content',
        ),
        migrations.AddField(
            model_name='chatmessage',
            name='_content',
            field=models.JSONField(default=''),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='chatmessage',
            name='_content_intel_class',
            field=models.CharField(default=' ', max_length=64),
            preserve_default=False,
        ),
    ]



from django_spire.metric.report.models import ReportRun
from django.contrib import admin


@admin.register(ReportRun)
class ReportRunAdmin(admin.ModelAdmin):
    list_display = ('id', 'report_key_stack_verbose', 'datetime')
    ordering = ('-datetime',)



from __future__ import annotations

from django.db.models import Choices


class MessageResponseType(Choices):
    REQUEST = 'request'
    RESPONSE = 'response'
    LOADING_RESPONSE = 'loading_response'



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.collection.tests.factories import create_test_collection
from django_spire.knowledge.entry.tests.factories import create_test_entry


class EntryModelTests(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.collection = create_test_collection()
        self.entry = create_test_entry(collection=self.collection)

    def test_str(self):
        assert str(self.entry) == self.entry.name

    def test_name_short_truncates(self):
        long_name = 'A' * 100
        entry = create_test_entry(name=long_name, collection=self.collection)
        assert len(entry.name_short) <= 35

    def test_name_short_no_truncate(self):
        short_name = 'Short'
        entry = create_test_entry(name=short_name, collection=self.collection)
        assert entry.name_short == short_name

    def test_top_level_collection(self):
        child_collection = create_test_collection(
            parent=self.collection,
            name='Child Collection'
        )
        entry = create_test_entry(collection=child_collection)
        assert entry.top_level_collection == self.collection

    def test_top_level_collection_when_already_top(self):
        assert self.entry.top_level_collection == self.collection

    def test_base_breadcrumb(self):
        breadcrumbs = self.entry.base_breadcrumb()
        assert breadcrumbs is not None



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from dandy.recorder import recorder_to_html_file

from django_spire.ai.chat.intelligence.decoders.intent_decoder import generate_intent_decoder
from django_spire.ai.chat.message_intel import BaseMessageIntel, DefaultMessageIntel
from django_spire.ai.decorators import log_ai_interaction_from_recorder

if TYPE_CHECKING:
    from dandy.llm.request.message import MessageHistory
    from django.core.handlers.wsgi import WSGIRequest


class BaseChatRouter(ABC):
    @abstractmethod
    def workflow(
        self,
        request: WSGIRequest,
        user_input: str,
        message_history: MessageHistory | None = None
    ) -> BaseMessageIntel:
        raise NotImplementedError

    @recorder_to_html_file('spire_ai_chat_workflow')
    def process(
        self,
        request: WSGIRequest,
        user_input: str,
        message_history: MessageHistory | None = None
    ) -> BaseMessageIntel:
        @log_ai_interaction_from_recorder(request.user)
        def run_workflow_process() -> BaseMessageIntel | None:
            return self.workflow(
                request=request,
                user_input=user_input,
                message_history=message_history,
            )

        message_intel = run_workflow_process()

        if not isinstance(message_intel, BaseMessageIntel):
            if message_intel is None:
                return DefaultMessageIntel(
                    text='I apologize, but I was unable to process your request.'
                )

            message = f'{self.__class__.__name__}.workflow must return an instance of a {BaseMessageIntel.__name__} sub class.'
            raise TypeError(message)

        return message_intel


class SpireChatRouter(BaseChatRouter):
    def _default_chat_callable(
        self,
        request: WSGIRequest,
        user_input: str,
        message_history: MessageHistory | None = None
    ) -> BaseMessageIntel:
        from django_spire.knowledge.intelligence.workflows.knowledge_workflow import (
            knowledge_search_workflow,
        )

        if request.user.has_perm('django_spire_knowledge.view_collection'):
            return knowledge_search_workflow(
                request=request,
                user_input=user_input,
                message_history=message_history
            )

        return DefaultMessageIntel(
            text='Sorry, I could not find any information on that.'
        )

    def workflow(
        self,
        request: WSGIRequest,
        user_input: str,
        message_history: MessageHistory | None = None
    ) -> BaseMessageIntel:
        intent_decoder = generate_intent_decoder(
            request=request,
            default_callable=self._default_chat_callable,
        )

        intent_process = intent_decoder.process(user_input, max_return_values=1)[0]

        return intent_process(
            request=request,
            user_input=user_input,
            message_history=message_history
        )



from __future__ import annotations

from django_spire.notification.app.models import AppNotification
from django_spire.notification.choices import NotificationTypeChoices
from django_spire.notification.email.models import EmailNotification
from django_spire.notification.push.models import PushNotification
from django_spire.notification.sms.models import SmsNotification


NOTIFICATION_TYPE_CHOICE_TO_MODEL_MAP: dict[NotificationTypeChoices, type] = {
    NotificationTypeChoices.APP: AppNotification,
    NotificationTypeChoices.EMAIL: EmailNotification,
    NotificationTypeChoices.PUSH: PushNotification,
    NotificationTypeChoices.SMS: SmsNotification
}



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.entry.version.block.choices import BlockTypeChoices
from django_spire.knowledge.entry.version.converters.markdown_converter import MarkdownConverter
from django_spire.knowledge.entry.version.tests.factories import create_test_entry_version


class MarkdownConverterTests(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.entry_version = create_test_entry_version()
        self.converter = MarkdownConverter(entry_version=self.entry_version)

    def test_convert_markdown_to_blocks_paragraph(self):
        markdown = 'This is a paragraph.'
        blocks = self.converter.convert_markdown_to_blocks(markdown)
        assert len(blocks) == 1
        assert blocks[0].type == BlockTypeChoices.TEXT

    def test_convert_markdown_to_blocks_heading(self):
        markdown = '# Heading 1'
        blocks = self.converter.convert_markdown_to_blocks(markdown)
        assert len(blocks) == 1
        assert blocks[0].type == BlockTypeChoices.HEADING

    def test_convert_markdown_to_blocks_multiple_headings(self):
        markdown = '# H1\n\n## H2\n\n### H3'
        blocks = self.converter.convert_markdown_to_blocks(markdown)
        heading_blocks = [b for b in blocks if b.type == BlockTypeChoices.HEADING]
        assert len(heading_blocks) == 3

    def test_convert_markdown_to_blocks_unordered_list(self):
        markdown = '- Item 1\n- Item 2\n- Item 3'
        blocks = self.converter.convert_markdown_to_blocks(markdown)
        list_blocks = [b for b in blocks if b.type == BlockTypeChoices.LIST]
        assert len(list_blocks) == 1

    def test_convert_markdown_to_blocks_ordered_list(self):
        markdown = '1. First\n2. Second\n3. Third'
        blocks = self.converter.convert_markdown_to_blocks(markdown)
        list_blocks = [b for b in blocks if b.type == BlockTypeChoices.LIST]
        assert len(list_blocks) == 1

    def test_convert_markdown_to_blocks_mixed_content(self):
        markdown = '# Title\n\nSome text.\n\n- List item'
        blocks = self.converter.convert_markdown_to_blocks(markdown)
        assert len(blocks) >= 3

    def test_convert_markdown_to_blocks_blank_lines(self):
        markdown = 'Para 1\n\n\n\nPara 2'
        blocks = self.converter.convert_markdown_to_blocks(markdown)
        assert len(blocks) >= 2

    def test_html_to_markdown(self):
        html = '<p>Hello <b>World</b></p>'
        result = MarkdownConverter.html_to_markdown(html)
        assert 'Hello' in result
        assert 'World' in result

    def test_convert_markdown_to_blocks_checklist(self):
        markdown = '- [ ] Todo\n- [x] Done'
        blocks = self.converter.convert_markdown_to_blocks(markdown)
        list_blocks = [b for b in blocks if b.type == BlockTypeChoices.LIST]
        assert len(list_blocks) == 1

    def test_convert_markdown_to_blocks_nested_list(self):
        markdown = '- Parent\n  - Child'
        blocks = self.converter.convert_markdown_to_blocks(markdown)
        list_blocks = [b for b in blocks if b.type == BlockTypeChoices.LIST]
        assert len(list_blocks) == 1

    def test_block_ordering(self):
        markdown = '# H1\n\nParagraph\n\n- List'
        blocks = self.converter.convert_markdown_to_blocks(markdown)
        for i, block in enumerate(blocks):
            assert block.order == i



from __future__ import annotations

from typing import TYPE_CHECKING

from dandy import Bot, Prompt

from django_spire.ai.chat.message_intel import BaseMessageIntel, DefaultMessageIntel
from django_spire.ai.chat.router import BaseChatRouter
from django_spire.conf import settings
from django_spire.knowledge.intelligence.workflows.knowledge_workflow import (
    NO_KNOWLEDGE_MESSAGE_INTEL,
    knowledge_search_workflow,
)

if TYPE_CHECKING:
    from dandy.llm.request.message import MessageHistory
    from django.core.handlers.wsgi import WSGIRequest


class KnowledgeSearchRouter(BaseChatRouter):
    def _default_chat_callable(
        self,
        request: WSGIRequest,
        user_input: str,
        message_history: MessageHistory | None = None
    ) -> BaseMessageIntel:
        persona_name = getattr(settings, 'DJANGO_SPIRE_AI_PERSONA_NAME', 'AI Assistant')

        system_prompt = (
            Prompt()
            .text(f'You are {persona_name}, a helpful AI assistant.')
            .line_break()
            .text('Important rules:')
            .list([
                f'You should always identify yourself as {persona_name}.',
                'Please never mention being Qwen, Claude, GPT, or any other model name.',
                'Be helpful, friendly, and professional.',
            ])
        )

        bot = Bot()
        bot.llm_role = system_prompt

        return bot.llm.prompt_to_intel(
            prompt=user_input,
            intel_class=DefaultMessageIntel,
            message_history=message_history,
        )

    def workflow(
        self,
        request: WSGIRequest,
        user_input: str,
        message_history: MessageHistory | None = None
    ) -> BaseMessageIntel:
        knowledge_result = knowledge_search_workflow(
            request=request,
            user_input=user_input,
            message_history=message_history
        )

        if knowledge_result == NO_KNOWLEDGE_MESSAGE_INTEL:
            return self._default_chat_callable(
                request=request,
                user_input=user_input,
                message_history=message_history
            )

        return knowledge_result



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.conf import settings
from django.urls import reverse

from django_spire.auth.controller.controller import AppAuthController
from django_spire.contrib import Breadcrumbs
from django_spire.contrib.generic_views import portal_views
from django_spire.core.utils import get_object_from_module_string
from django_spire.metric.report.models import ReportRun
from django_spire.metric.report.registry import ReportRegistry

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from django.template.response import TemplateResponse


@AppAuthController('report').permission_required('can_view')
def report_view(request: WSGIRequest) -> TemplateResponse:
    breadcrumbs = Breadcrumbs()

    breadcrumbs.add_breadcrumb(
        'Reports',
        reverse('django_spire:metric:report:page:report'),
    )

    page_report_registry = ReportRegistry()

    for report_registry in settings.DJANGO_SPIRE_REPORT_REGISTRIES:
        report_registry_class = get_object_from_module_string(
            report_registry
        )

        page_report_registry.add_registry(
            report_registry_class()
        )

    context_data = dict()

    context_data['registry'] = page_report_registry

    if request.GET:
        report_key_stack = request.GET.get('report_key_stack', None)

        if report_key_stack:
            report = page_report_registry.get_report_from_key_stack(report_key_stack)

            if report:
                for key in report_key_stack.split('|'):
                    breadcrumbs.add_breadcrumb(
                        key,
                    )

                context_data['report_run_arguments'] = report.run_arguments

                context_data['report_run_arguments_values'] = {}

                for argument in context_data['report_run_arguments']:
                    if context_data['report_run_arguments'][argument]['annotation'] == 'bool':
                        get_request_value = request.GET.get(argument, False)

                    elif context_data['report_run_arguments'][argument]['annotation'] == 'date':
                        date_str = request.GET.get(argument, None)

                        if date_str:
                            get_request_value = datetime.strptime(date_str, '%Y-%m-%d').date()
                        else:
                            get_request_value = date_str


                    elif context_data['report_run_arguments'][argument]['annotation'] == 'datetime':
                        datetime_str = request.GET.get(argument, None)

                        if datetime_str:
                            get_request_value = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S')
                        else:
                            get_request_value = datetime_str

                    elif context_data['report_run_arguments'][argument]['annotation'] == 'multi_select':
                        get_request_value = request.GET.getlist(argument, [])

                    else:
                        value = request.GET.get(argument, None)

                        if value:
                            get_request_value = context_data['report_run_arguments'][argument]['annotation_class'](value)
                        else:
                            get_request_value = value

                    context_data['report_run_arguments_values'][argument] = get_request_value

                if request.GET.get('report_should_run', 'false').lower() == 'true':
                    for argument, value in context_data['report_run_arguments_values'].items():
                        if value is None:
                            break
                    else:
                        ReportRun.objects.create(
                            report_key_stack=report_key_stack,
                        )
                        report.run(**context_data['report_run_arguments_values'])

                context_data['report'] = report
                context_data['report_run_count'] = ReportRun.objects.run_count(report_key_stack)

    else:
        top_ten_report_runs = [
            {
                **report_run,
                'report_key_stack_verbose': report_run['report_key_stack'].replace('|', ' > '),
            }
            for report_run in
            ReportRun.objects.by_top_ten()
        ]

        context_data['top_ten_report_runs'] = top_ten_report_runs

    return portal_views.template_view(
        request,
        page_title='Reports',
        page_description='More Reporting Info',
        breadcrumbs=breadcrumbs,
        context_data=context_data,
        template='django_spire/metric/report/page/report_page.html'
    )



import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass
from dataclasses import field
from typing import Callable, Any

from django_spire.metric.report.enums import ColumnType
from django_spire.metric.report.helper import Helper
from django_spire.metric.report.tools import get_text_alignment_css_class


@dataclass
class ReportColumn:
    title: str
    sub_title: str | None = None
    type: ColumnType = ColumnType.TEXT
    sub_type: ColumnType = ColumnType.TEXT

    def css_class(self) -> str:
        return get_text_alignment_css_class(self.type)


@dataclass
class ReportCell:
    value: Any
    sub_value: Any = None
    type: ColumnType = ColumnType.TEXT
    sub_type: ColumnType = ColumnType.TEXT

    def css_class(self) -> str:
        return get_text_alignment_css_class(self.type)

    @staticmethod
    def cell_value_verbose(value, cell_type):
        if cell_type == ColumnType.DOLLAR:
            return f"${float(value):,.0f}"
        elif cell_type == ColumnType.DOLLAR_1:
            return f"${float(value):,.1f}%"
        elif cell_type == ColumnType.DOLLAR_2:
            return f"${float(value):,.2f}%"
        elif cell_type == ColumnType.DOLLAR_3:
            return f"${float(value):,.3f}%"

        elif cell_type == ColumnType.PERCENT:
            return f"{float(value):,.0f}%"
        elif cell_type == ColumnType.PERCENT_1:
            return f"{float(value):,.1f}%"
        elif cell_type == ColumnType.PERCENT_2:
            return f"{float(value):,.2f}%"
        elif cell_type == ColumnType.PERCENT_3:
            return f"{float(value):,.3f}%"

        elif cell_type == ColumnType.NUMBER:
            return f"{float(value):,.0f}"
        elif cell_type == ColumnType.NUMBER_1:
            return f"{float(value):,.1f}"
        elif cell_type == ColumnType.NUMBER_2:
            return f"{float(value):,.2f}"
        elif cell_type == ColumnType.NUMBER_3:
            return f"{float(value):,.3f}"

        return str(value)

    def value_verbose(self):
        return self.cell_value_verbose(self.value, self.type)

    def sub_value_verbose(self):
        return self.cell_value_verbose(self.sub_value, self.sub_type)


@dataclass
class ReportRow:
    cells: list[ReportCell] = field(default_factory=list)
    bold: bool = False
    page_break: bool = False
    span_all_columns: bool = False
    table_break: bool = False
    border_top: bool = False
    border_bottom: bool = False


class BaseReport(ABC):
    title: str
    description: str | None = None
    is_financially_accurate: bool = False
    ColumnType: type[ColumnType] = ColumnType
    helper: Helper = Helper()

    def __init__(self):
        if not self.title:
            message = 'Report title is required'
            raise ValueError(message)

        self.columns: list[ReportColumn] = []
        self.rows: list[ReportRow] = []

    @property
    def column_count(self) -> int:
        return len(self.columns)

    @property
    def is_ready(self):
        return len(self.columns) > 0

    @property
    def run_arguments(self) -> dict[str, dict[str, str]]:
        arguments = {}
        signature = inspect.signature(self.run)

        for name, param in signature.parameters.items():
            arguments[name] = {}
            arguments[name]['default'] = param.default
            arguments[name]['annotation_class'] = param.annotation

            choices_method = getattr(self, f'{name}_choices', None)

            if choices_method and isinstance(choices_method, Callable):
                choices = tuple(choices_method())

                self.validate_choices(tuple(choices))

                arguments[name]['choices'] = choices
                if param.annotation.__name__ == 'list':
                    arguments[name]['annotation'] = 'multi_select'
                else:
                    arguments[name]['annotation'] = 'select'
            else:
                arguments[name]['annotation'] = param.annotation.__name__

        return arguments

    @abstractmethod
    def run(self, **kwargs: Any):
        raise NotImplementedError

    def add_blank_row(
            self,
            text: str = '',
            page_break: bool = False,
            border_top: bool = False,
            border_bottom: bool = False
    ):
        self.add_row(
            cell_values=[
                text
            ],
            span_all_columns=True,
            page_break=page_break,
            border_top=border_top,
            border_bottom=border_bottom,
        )

    def add_column(
            self,
            title: str,
            sub_title: str | None = None,
            type: ColumnType = ColumnType.TEXT,
            sub_type: ColumnType = ColumnType.TEXT,

    ):
        self.columns.append(
            ReportColumn(title=title, sub_title=sub_title, type=type, sub_type=sub_type)
        )

    def add_divider_row(
            self,
            title: str,
            description: str | None = None,
            page_break: bool = False,
            border_bottom: bool = True,
    ):
        self.add_row(
            cell_values=[title],
            cell_sub_values=[description] if description else None,
            bold=True,
            page_break=page_break,
            span_all_columns=True,
            border_bottom=border_bottom,
        )

    def add_footer_row(
            self,
            cell_values: list[Any],
            cell_sub_values: list[Any] | None = None,
            border_top: bool = True,
    ):
        self.add_row(
            cell_values=cell_values,
            cell_sub_values=cell_sub_values,
            bold=True,
            border_top=border_top,
        )

    def add_row(
            self,
            cell_values: list[Any],
            cell_sub_values: list[Any] | None = None,
            bold: bool = False,
            page_break: bool = False,
            span_all_columns: bool = False,
            table_break: bool = False,
            border_top: bool = False,
            border_bottom: bool = False,
    ):
        if span_all_columns or table_break:
            if len(cell_values) > 1:
                message = 'Cannot span all columns or have a table break with more than one cell value'
                raise ValueError(message)

        elif len(cell_values) != len(self.columns) or (
                cell_sub_values is not None and len(cell_sub_values) != len(self.columns)):
            message = f'Number of cell values ({len(cell_values)}) and sub values ({len(cell_sub_values) if cell_sub_values else "None"}) must match number of columns: {len(self.columns)}'
            raise ValueError(message)

        self.rows.append(
            ReportRow(
                cells=[
                    ReportCell(
                        value=cell_values[i],
                        sub_value=cell_sub_values[i] if cell_sub_values else None,
                        type=self.columns[i].type,
                        sub_type=self.columns[i].sub_type
                    )
                    for i in range(len(cell_values))
                ],
                bold=bold,
                page_break=page_break,
                span_all_columns=span_all_columns,
                table_break=table_break,
                border_top=border_top,
                border_bottom=border_bottom,
            )
        )

    @staticmethod
    def validate_choices(choices: tuple):
        if not isinstance(choices, tuple):
            raise TypeError(f'choices must be a tuple not {type(choices)}')
        if not all(isinstance(item, tuple) and len(item) == 2 for item in choices):
            raise ValueError('choices must contain tuples of length 2')

    def to_markdown(self) -> str:
        markdown = ''

        for column in self.columns:
            markdown += f'| {column.title} '

        markdown += '|\n'

        for column in self.columns:
            markdown += '| ' + '-' * len(column.title) + ' '

        markdown += '|\n'

        for row in self.rows:
            if row.span_all_columns:
                markdown += f'| {row.cells[0].value}' + '|' * len(self.columns) + '\n'
                continue

            else:
                for cell in row.cells:
                    markdown += f'| {cell.value_verbose()} '

            markdown += '|\n'

        return markdown



from __future__ import annotations

from django.urls import reverse

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.help_desk.models import HelpDeskTicket
from django_spire.help_desk.tests.factories import (
    create_test_helpdesk_ticket,
    create_test_helpdesk_ticket_data,
)


class HelpDeskFormViewsTestCase(BaseTestCase):
    def test_helpdesk_ticket_form_create_view(self):
        ticket_data = create_test_helpdesk_ticket_data()

        response = self.client.post(
            reverse('django_spire:help_desk:form:create'),
            data=ticket_data,
        )

        ticket = HelpDeskTicket.objects.first()

        assert response.status_code == 302
        assert response.url == reverse('django_spire:help_desk:page:list')

        assert ticket.priority == ticket_data['priority']
        assert ticket.purpose == ticket_data['purpose']
        assert ticket.status == ticket_data['status']
        assert ticket.description == ticket_data['description']

    def test_helpdesk_ticket_form_update_view(self):
        test_ticket = create_test_helpdesk_ticket()

        updated_ticket_data = create_test_helpdesk_ticket_data(
            priority='high',
            purpose='comp',
            status='prog',
            description='This is an updated test ticket.',
        )

        response = self.client.post(
            data=updated_ticket_data,
            path=reverse(
                'django_spire:help_desk:form:update',
                kwargs={'pk': test_ticket.pk}
            ),
        )

        assert response.status_code == 302
        assert response.url == reverse('django_spire:help_desk:page:list')

        test_ticket.refresh_from_db()
        assert test_ticket.priority == updated_ticket_data['priority']
        assert test_ticket.purpose == updated_ticket_data['purpose']
        assert test_ticket.status == updated_ticket_data['status']
        assert test_ticket.description == updated_ticket_data['description']

    def test_helpdesk_ticket_form_create_view_invalid_data(self):
        invalid_data = {
            'priority': '',
            'purpose': '',
            'description': '',
        }

        response = self.client.post(
            reverse('django_spire:help_desk:form:create'),
            data=invalid_data,
        )

        assert response.status_code == 200
        assert HelpDeskTicket.objects.count() == 0

    def test_helpdesk_ticket_form_create_view_sets_created_by(self):
        ticket_data = create_test_helpdesk_ticket_data()

        self.client.post(
            reverse('django_spire:help_desk:form:create'),
            data=ticket_data,
        )

        ticket = HelpDeskTicket.objects.first()
        assert ticket.created_by == self.super_user



from __future__ import annotations

import hashlib

from dataclasses import dataclass

from django.core.cache import cache

from django_spire.knowledge.intelligence.bots.search_preprocessing_bot import (
    PreprocessedQueryIntel,
    SearchPreprocessingBot,
)


PREPROCESSING_CACHE_TIMEOUT = 3600


def _get_cache_key(query: str) -> str:
    normalized = query.lower().strip()
    query_hash = hashlib.sha256(normalized.encode()).hexdigest()
    return f'knowledge_search_preprocess:{query_hash}'


@dataclass
class PreprocessedSearchQuery:
    corrected_query: str
    expanded_terms: list[str]
    meaningful_words: list[str]
    original_query: str
    search_phrases: list[str]

    @property
    def all_search_terms(self) -> set[str]:
        terms = set(self.meaningful_words)
        terms.update(self.expanded_terms)

        if not terms:
            terms.add(self.corrected_query or self.original_query)

        return terms

    @property
    def primary_search_query(self) -> str:
        if self.meaningful_words:
            return ' '.join(self.meaningful_words)

        return self.corrected_query or self.original_query


def preprocess_search_query(
    query: str,
    use_cache: bool = True,
    llm_temperature: float = 0.3
) -> PreprocessedSearchQuery:
    if use_cache:
        cache_key = _get_cache_key(query)
        cached = cache.get(cache_key)

        if cached:
            return cached

    intel: PreprocessedQueryIntel = (
        SearchPreprocessingBot(llm_temperature=llm_temperature)
        .process(query=query)
    )

    result = PreprocessedSearchQuery(
        corrected_query=intel.corrected_query,
        expanded_terms=intel.expanded_terms,
        meaningful_words=intel.meaningful_words,
        original_query=query,
        search_phrases=intel.search_phrases,
    )

    if use_cache:
        cache.set(cache_key, result, PREPROCESSING_CACHE_TIMEOUT)

    return result



from __future__ import annotations

from unittest import TestCase

from django_spire.core.tag.tools import (
    get_matching_a_percentage_from_tag_sets,
    get_matching_b_percentage_from_tag_sets,
    get_matching_count_from_tag_sets,
    get_matching_tag_set_from_tag_sets,
    get_score_percentage_from_tag_set_weighted,
    simplify_and_weight_tag_set_to_dict,
    simplify_tag_set,
    simplify_tag_set_to_list,
)


class TestGetMatchingTagSetFromTagSets(TestCase):
    def test_full_intersection(self) -> None:
        tag_set_a = {'apple', 'banana', 'orange'}
        tag_set_b = {'apple', 'banana', 'orange'}

        result = get_matching_tag_set_from_tag_sets(tag_set_a, tag_set_b)

        assert result == {'apple', 'banana', 'orange'}

    def test_no_intersection(self) -> None:
        tag_set_a = {'apple', 'banana', 'orange'}
        tag_set_b = {'grape', 'kiwi', 'pear'}

        result = get_matching_tag_set_from_tag_sets(tag_set_a, tag_set_b)

        assert result == set()

    def test_partial_intersection(self) -> None:
        tag_set_a = {'apple', 'banana', 'orange'}
        tag_set_b = {'apple', 'grape', 'kiwi'}

        result = get_matching_tag_set_from_tag_sets(tag_set_a, tag_set_b)

        assert result == {'apple'}


class TestGetMatchingCountFromTagSets(TestCase):
    def test_full_match(self) -> None:
        tag_set_a = {'apple', 'banana'}
        tag_set_b = {'apple', 'banana'}

        assert get_matching_count_from_tag_sets(tag_set_a, tag_set_b) == 2

    def test_no_match(self) -> None:
        tag_set_a = {'apple', 'banana'}
        tag_set_b = {'grape', 'kiwi'}

        assert get_matching_count_from_tag_sets(tag_set_a, tag_set_b) == 0

    def test_partial_match(self) -> None:
        tag_set_a = {'apple', 'banana', 'orange'}
        tag_set_b = {'apple', 'grape'}

        assert get_matching_count_from_tag_sets(tag_set_a, tag_set_b) == 1


class TestGetMatchingAPercentageFromTagSets(TestCase):
    def test_empty_tag_set_a(self) -> None:
        tag_set_a = set()
        tag_set_b = {'apple', 'banana'}

        assert get_matching_a_percentage_from_tag_sets(tag_set_a, tag_set_b) == 0.0

    def test_full_match(self) -> None:
        tag_set_a = {'apple', 'banana'}
        tag_set_b = {'apple', 'banana', 'orange'}

        assert get_matching_a_percentage_from_tag_sets(tag_set_a, tag_set_b) == 1.0

    def test_partial_match(self) -> None:
        tag_set_a = {'apple', 'banana', 'orange'}
        tag_set_b = {'apple'}

        assert abs(get_matching_a_percentage_from_tag_sets(tag_set_a, tag_set_b) - 0.333) < 0.01


class TestGetMatchingBPercentageFromTagSets(TestCase):
    def test_empty_tag_set_b(self) -> None:
        tag_set_a = {'apple', 'banana'}
        tag_set_b = set()

        assert get_matching_b_percentage_from_tag_sets(tag_set_a, tag_set_b) == 0.0

    def test_full_match(self) -> None:
        tag_set_a = {'apple', 'banana', 'orange'}
        tag_set_b = {'apple', 'banana'}

        assert get_matching_b_percentage_from_tag_sets(tag_set_a, tag_set_b) == 1.0

    def test_partial_match(self) -> None:
        tag_set_a = {'apple'}
        tag_set_b = {'apple', 'banana', 'orange'}

        assert abs(get_matching_b_percentage_from_tag_sets(tag_set_a, tag_set_b) - 0.333) < 0.01


class TestSimplifyTagSet(TestCase):
    def test_no_hyphens(self) -> None:
        tag_set = {'apple', 'banana', 'orange'}

        result = simplify_tag_set(tag_set)

        assert result == {'apple', 'banana', 'orange'}

    def test_with_hyphens(self) -> None:
        tag_set = {'machine-learning', 'data-science'}

        result = simplify_tag_set(tag_set)

        assert 'machine' in result
        assert 'learning' in result
        assert 'data' in result
        assert 'science' in result


class TestSimplifyTagSetToList(TestCase):
    def test_mixed_tags(self) -> None:
        tag_set = {'apple', 'machine-learning'}

        result = simplify_tag_set_to_list(tag_set)

        assert 'apple' in result
        assert 'machine' in result
        assert 'learning' in result

    def test_with_hyphens(self) -> None:
        tag_set = {'machine-learning'}

        result = simplify_tag_set_to_list(tag_set)

        assert result == ['machine', 'learning']


class TestSimplifyAndWeightTagSetToDict(TestCase):
    def test_duplicate_words(self) -> None:
        tag_set = {'machine-learning', 'deep-learning'}

        result = simplify_and_weight_tag_set_to_dict(tag_set)

        assert result['learning'] == 2
        assert result['machine'] == 1
        assert result['deep'] == 1

    def test_simple_tags(self) -> None:
        tag_set = {'apple', 'banana'}

        result = simplify_and_weight_tag_set_to_dict(tag_set)

        assert result['apple'] == 1
        assert result['banana'] == 1

    def test_sorted_by_weight(self) -> None:
        tag_set = {'machine-learning', 'deep-learning', 'learning'}

        result = simplify_and_weight_tag_set_to_dict(tag_set)
        keys = list(result.keys())

        assert keys[0] == 'learning'
        assert result['learning'] == 3


class TestGetScorePercentageFromTagSetWeighted(TestCase):
    def test_empty_reference(self) -> None:
        tag_set_actual = {'apple', 'banana'}
        tag_set_reference = set()

        result = get_score_percentage_from_tag_set_weighted(tag_set_actual, tag_set_reference)

        assert result == 0.0

    def test_no_match(self) -> None:
        tag_set_actual = {'grape', 'kiwi'}
        tag_set_reference = {'apple', 'banana'}

        result = get_score_percentage_from_tag_set_weighted(tag_set_actual, tag_set_reference)

        assert result == 0.0

    def test_with_weighted_match(self) -> None:
        tag_set_actual = {'machine-learning', 'learning'}
        tag_set_reference = {'machine-learning', 'deep-learning'}

        result = get_score_percentage_from_tag_set_weighted(tag_set_actual, tag_set_reference)

        assert result > 0



from __future__ import annotations

import json

from typing import TYPE_CHECKING

from django.http import JsonResponse
from django.shortcuts import get_object_or_404

from django_spire.auth.controller.controller import AppAuthController
from django_spire.core.decorators import valid_ajax_request_required
from django_spire.knowledge.entry.version.models import EntryVersion

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


@valid_ajax_request_required
@AppAuthController('knowledge').permission_required('can_change')
def update_blocks_view(request: WSGIRequest, pk: int) -> JsonResponse:
    entry_version = get_object_or_404(EntryVersion.objects.prefetch_blocks(), pk=pk)

    block_data_list = json.loads(request.body.decode('utf-8'))

    entry_version.services.processor.add_update_delete_blocks(block_data_list)

    return JsonResponse({'type': 'success'})

@valid_ajax_request_required
@AppAuthController('knowledge').permission_required('can_change')
def update_entry_from_version_view(request: WSGIRequest, pk: int) -> JsonResponse:
    entry_version = get_object_or_404(EntryVersion, pk=pk)

    entry_version.entry.services.tag.process_and_set_tags()

    return JsonResponse({'type': 'success'})



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.sms.choices import SmsMediaContentTypeChoices


class SmsMediaContentTypeChoicesTests(BaseTestCase):
    def test_png_value(self):
        assert SmsMediaContentTypeChoices.PNG == 'image/png'

    def test_jpeg_value(self):
        assert SmsMediaContentTypeChoices.JPEG == 'image/jpeg'

    def test_choices_count(self):
        assert len(SmsMediaContentTypeChoices.choices) == 2



from __future__ import annotations

from django.urls import include, path


app_name = 'report'

urlpatterns = [
    path('page/', include('django_spire.metric.report.urls.page_urls', namespace='page')),
]



from __future__ import annotations

import os
import pytest

from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Page
    from pytest_django.plugin import _LiveServer


def pytest_configure(config: Any) -> None:
    os.environ['DJANGO_ALLOW_ASYNC_UNSAFE'] = 'true'


@pytest.fixture(scope='session')
def browser_context_args(browser_context_args: dict[str, Any]) -> dict[str, Any]:
    return {
        **browser_context_args,
        'no_viewport': True,
        'viewport': None,
    }


@pytest.fixture(scope='session')
def browser_type_launch_args(browser_type_launch_args: dict[str, Any]) -> dict[str, Any]:
    return {
        **browser_type_launch_args,
        'args': ['--start-maximized'],
    }


@pytest.fixture
def authenticated_page(page: Page, live_server: _LiveServer, transactional_db: None) -> Page:
    from django.contrib.auth import get_user_model

    User = get_user_model()

    User.objects.create_user(
        username='testuser',
        password='testpass123',
        is_staff=True,
        is_superuser=True
    )

    page.goto(f'{live_server.url}/admin/login/')
    page.fill('input[name="username"]', 'testuser')
    page.fill('input[name="password"]', 'testpass123')
    page.click('input[type="submit"]')
    page.wait_for_url(f'{live_server.url}/admin/')

    return page



from __future__ import annotations

from django.contrib.auth.models import Permission, User
from django.test import RequestFactory

from django_spire.ai.chat.auth.controller import BaseAiChatAuthController
from django_spire.core.tests.test_cases import BaseTestCase


class BaseAiChatAuthControllerTests(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.factory = RequestFactory()
        self.request = self.factory.get('/')
        self.request.user = self.super_user

    def test_can_delete_with_permission(self) -> None:
        permission = Permission.objects.get(codename='delete_chat')
        self.super_user.user_permissions.add(permission)

        self.super_user = type(self.super_user).objects.get(pk=self.super_user.pk)
        self.request.user = self.super_user

        controller = BaseAiChatAuthController(self.request)

        assert controller.can_delete() is True

    def test_can_delete_without_permission(self) -> None:
        regular_user = User.objects.create_user(
            username='regular_user',
            password='testpass123'  # noqa: S106
        )

        request = self.factory.get('/')
        request.user = regular_user

        controller = BaseAiChatAuthController(request)

        assert controller.can_delete() is False

    def test_controller_has_request(self) -> None:
        controller = BaseAiChatAuthController(self.request)

        assert controller.request == self.request



from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey
from django.db.models import QuerySet, Field, ForeignObjectRel, ForeignKey


def filter_by_lookup_map(
    queryset: QuerySet,
    lookup_map: dict,
    data: dict,
    extra_filters: dict | None = None
):
    """
    Filters a given queryset based on a lookup map and provided data. Additional filters
    can also be applied if provided.

    :param queryset (QuerySet): The queryset to be filtered.
    :param lookup_map (dict): A dictionary mapping input data keys to filtering fields in
        the queryset.
    :param data (dict): A dictionary containing key-value pairs to filter the queryset.
    :param extra_filters (list | None): Optional. A list of extra positional filter
        arguments to be applied. Defaults to None.

    :returns: QuerySet: The filtered queryset.
    """
    if extra_filters is None:
        extra_filters = {}

    lookup_kwargs = {
        lookup_map[k]: v
        for k, v in data.items()
        if k in lookup_map and v not in (None, "", [])
    } | extra_filters

    return queryset.filter(**lookup_kwargs)


def _get_kwarg_name_for_filter_field(
    field: Field[Any, Any] | ForeignObjectRel | GenericForeignKey,
    val: Any
) -> str:
    """
    Determines the appropriate keyword argument name for filtering based on the field type
    and value type.

    :param field (Field | ForeignObjectRel | GenericForeignKey): The model field to generate
        the filter keyword argument for.
    :param val (Any): The value to be used for filtering.

    :returns: str: The keyword argument string suitable for filtering.
    """
    if isinstance(val, Sequence):
        if isinstance(field, ForeignKey):
            return f'{field.name}_id__in'
        return f'{field.name}__in'
    else:
        return field.name


def filter_by_model_fields(queryset: QuerySet, data: dict) -> QuerySet:
    """
    Filters a given queryset based on the queryset's model fields and provided data.

    :param queryset (QuerySet): The queryset to be filtered.
    :param data (dict): A dictionary containing key-value pairs to filter the queryset.

    :returns: QuerySet: The filtered queryset.
    """
    model_fields = [f for f in queryset.model._meta.get_fields()]

    lookup_kwargs = {
        _get_kwarg_name_for_filter_field(
            field=model_field,
            val=data[model_field.name]
        ): data[model_field.name]
        for model_field in model_fields
        if model_field.name in data and data[model_field.name] not in (None, "", [])
    }

    return queryset.filter(**lookup_kwargs)



from __future__ import annotations

import pytest

from django.test import TestCase

from django_spire.core.utils import (
    get_callable_from_module_string_and_validate_arguments,
    get_object_from_module_string
)


class TestGetObjectFromModuleString(TestCase):
    def test_import_class(self) -> None:
        obj = get_object_from_module_string('django.test.TestCase')
        assert obj is TestCase

    def test_import_function(self) -> None:
        obj = get_object_from_module_string('os.path.join')
        assert callable(obj)

    def test_import_invalid_module(self) -> None:
        with pytest.raises(ImportError, match='Could not import module'):
            get_object_from_module_string('nonexistent.module.Object')

    def test_import_module_constant(self) -> None:
        obj = get_object_from_module_string('os.sep')
        assert isinstance(obj, str)


class TestGetCallableFromModuleStringAndValidateArguments(TestCase):
    def test_callable_with_missing_argument(self) -> None:
        with pytest.raises(TypeError, match='missing required argument'):
            get_callable_from_module_string_and_validate_arguments(
                'json.dumps',
                ['nonexistent_arg']
            )

    def test_callable_with_valid_arguments(self) -> None:
        callable_ = get_callable_from_module_string_and_validate_arguments(
            'json.dumps',
            ['obj']
        )
        assert callable(callable_)

    def test_invalid_module(self) -> None:
        with pytest.raises(ImportError, match='Could not import module'):
            get_callable_from_module_string_and_validate_arguments(
                'nonexistent.module.func',
                []
            )

    def test_non_callable_object(self) -> None:
        with pytest.raises(TypeError, match='is not callable'):
            get_callable_from_module_string_and_validate_arguments(
                'os.sep',
                []
            )



from __future__ import annotations

from typing import TYPE_CHECKING

from django.db.models import QuerySet, Value, When, Case, BooleanField

from django_spire.history.querysets import HistoryQuerySet
from django_spire.notification.choices import NotificationStatusChoices
from django_spire.notification.querysets import NotificationContentObjectQuerySet

if TYPE_CHECKING:
    from django.contrib.auth.models import User


class EmailNotificationQuerySet(HistoryQuerySet, NotificationContentObjectQuerySet):
    def annotate_is_viewed_by_user(self, user: User) -> QuerySet:
        return self.by_user(user=user).annotate(viewed=Case(
            When(views__user=user, then=Value(True)),
            default=Value(False),
            output_field=BooleanField()
        ))

    def by_user(self, user: User) -> QuerySet:
        return self.filter(notification__user=user)

    def by_users(self, users: list[User]):
        return self.filter(notification__user__in=users)

    def exclude_viewed_by_user(self, user: User) -> QuerySet:
        return self.by_user(user=user).exclude(views__user=user)

    def is_sent(self) -> QuerySet:
        return self.filter(notification__status=NotificationStatusChoices.SENT)
